"""
BitBox02
********
"""

from typing import (
    cast,
    Any,
    Callable,
    Dict,
    Optional,
    Mapping,
    Union,
    Tuple,
    List,
    Sequence,
    TypeVar,
)
import builtins
import sys
from functools import wraps

import base58

from ..descriptor import MultisigDescriptor
from ..hwwclient import HardwareWalletClient
from ..key import ExtendedKey
from .._script import (
    is_p2pkh,
    is_p2wpkh,
    is_p2wsh,
    parse_multisig,
)
from ..psbt import PSBT
from ..tx import (
    CTxOut,
)
from .._serialize import (
    ser_uint256,
    ser_sig_der,
)
from ..errors import (
    HWWError,
    ActionCanceledError,
    BadArgumentError,
    DeviceNotReadyError,
    UnavailableActionError,
    DEVICE_NOT_INITIALIZED,
    handle_errors,
    common_err_msgs,
)
from ..key import (
    KeyOriginInfo,
    parse_path,
)
from ..common import (
    AddressType,
    Chain,
)

import hid

from bitbox02 import util
from bitbox02 import bitbox02
from bitbox02.communication import (
    devices,
    u2fhid,
    FirmwareVersionOutdatedException,
    Bitbox02Exception,
    UserAbortException,
    HARDENED,
    ERR_GENERIC,
)

from bitbox02.communication.bitbox_api_protocol import (
    Platform,
    BitBox02Edition,
    BitBoxNoiseConfig,
)


class BitBox02Error(UnavailableActionError):
    def __init__(self, msg: str):
        """
        BitBox02 unexpected error. The BitBox02 does not return give granular error messages,
        so we give hints to as what could be wrong.
        """
        msg = "Input error: {}. A keypath might be invalid. Supported keypaths are: ".format(
            msg
        )
        msg += "m/49'/0'/<account'> for p2wpkh-p2sh; "
        msg += "m/84'/0'/<account'> for p2wpkh; "
        msg += "m/48'/0'/<account'>/2' for p2wsh multisig; "
        msg += "m/48'/0'/<account'>/1' for p2wsh-p2sh multisig; "
        msg += "m/48'/0'/<account'>' for any supported multisig; "
        msg += "account can be between 0' and 99'; "
        msg += "For address keypaths, append /0/<address index> for a receive and /1/<change index> for a change address."
        super().__init__(msg)


ERR_INVALID_INPUT = 101

PURPOSE_P2WPKH_P2SH = 49 + HARDENED
PURPOSE_P2WPKH = 84 + HARDENED
PURPOSE_MULTISIG_P2WSH = 48 + HARDENED

# External GUI tools using hwi.py as a command line tool to integrate hardware wallets usually do
# not have an actual terminal for IO.
_using_external_gui = not sys.stdout.isatty()
if _using_external_gui:
    _unpaired_errmsg = "Device not paired yet. Please pair using the BitBoxApp, then close the BitBoxApp and try again."
else:
    _unpaired_errmsg = "Device not paired yet. Please use any subcommand to pair"


class SilentNoiseConfig(util.BitBoxAppNoiseConfig):
    """
    Used during `enumerate()`. Raises an exception if the device is unpaired.
    Attestation check is silent.

    Rationale: enumerate() should not show any dialogs.
    """

    def show_pairing(self, code: str, device_response: Callable[[], bool]) -> bool:
        raise DeviceNotReadyError(_unpaired_errmsg)

    def attestation_check(self, result: bool) -> None:
        pass


class CLINoiseConfig(util.BitBoxAppNoiseConfig):
    """ Noise pairing and attestation check handling in the terminal (stdin/stdout) """

    def show_pairing(self, code: str, device_response: Callable[[], bool]) -> bool:
        if _using_external_gui:
            # The user can't see the pairing in the terminal. The
            # output format is also not appropriate for parsing by
            # external tools doing inter process communication using
            # stdin/stdout. For now, we direct the user to pair in the
            # BitBoxApp instead.
            raise DeviceNotReadyError(_unpaired_errmsg)

        print("Please compare and confirm the pairing code on your BitBox02:")
        print(code)
        if not device_response():
            return False
        return input("Accept pairing? [y]/n: ").strip() != "n"

    def attestation_check(self, result: bool) -> None:
        if result:
            sys.stderr.write("BitBox02 attestation check PASSED\n")
        else:
            sys.stderr.write("BitBox02 attestation check FAILED\n")
            sys.stderr.write(
                "Your BitBox02 might not be genuine. Please contact support@shiftcrypto.ch if the problem persists.\n"
            )


def _keypath_hardened_prefix(keypath: Sequence[int]) -> Sequence[int]:
    for i, e in builtins.enumerate(keypath):
        if e & HARDENED == 0:
            return keypath[:i]
    return keypath


def _xpubs_equal_ignoring_version(xpub1: bytes, xpub2: bytes) -> bool:
    """
    Xpubs: 78 bytes. Returns true if the xpubs are equal, ignoring the 4 byte version.
    The version is not important and allows compatibility with Electrum, which exports PSBTs with
    xpubs using Electrum-style xpub versions.
    """
    return xpub1[4:] == xpub2[4:]


def enumerate(password: str = "") -> List[Dict[str, object]]:
    """
    Enumerate all BitBox02 devices. Bootloaders excluded.
    """
    result = []
    for device_info in devices.get_any_bitbox02s():
        path = device_info["path"].decode()
        client = Bitbox02Client(path)
        client.set_noise_config(SilentNoiseConfig())
        d_data: Dict[str, object] = {}
        bb02 = None
        with handle_errors(common_err_msgs["enumerate"], d_data):
            bb02 = client.init(expect_initialized=None)
        version, platform, edition, unlocked = bitbox02.BitBox02.get_info(
            client.transport
        )
        if platform != Platform.BITBOX02:
            client.close()
            continue
        if edition not in (BitBox02Edition.MULTI, BitBox02Edition.BTCONLY):
            client.close()
            continue

        assert isinstance(edition, BitBox02Edition)

        d_data.update(
            {
                "type": "bitbox02",
                "path": path,
                "model": {
                    BitBox02Edition.MULTI: "bitbox02_multi",
                    BitBox02Edition.BTCONLY: "bitbox02_btconly",
                }[edition],
                "needs_pin_sent": False,
                "needs_passphrase_sent": False,
            }
        )

        if bb02 is not None:
            with handle_errors(common_err_msgs["enumerate"], d_data):
                if not bb02.device_info()["initialized"]:
                    raise DeviceNotReadyError(
                        "BitBox02 is not initialized. Please initialize it using the BitBoxApp."
                    )
                elif not unlocked:
                    raise DeviceNotReadyError(
                        "Please load wallet to unlock."
                        if _using_external_gui
                        else "Please use any subcommand to unlock"
                    )
                d_data["fingerprint"] = client.get_master_fingerprint().hex()

        result.append(d_data)

        client.close()
    return result


T = TypeVar("T", bound=Callable[..., Any])


def bitbox02_exception(f: T) -> T:
    """
    Maps bitbox02 library exceptions into a HWI exceptions.
    """

    @wraps(f)
    def func(*args, **kwargs):  # type: ignore
        """ Wraps f, mapping exceptions. """
        try:
            return f(*args, **kwargs)
        except UserAbortException:
            raise ActionCanceledError("{} canceled".format(f.__name__))
        except Bitbox02Exception as exc:
            if exc.code in (ERR_GENERIC, ERR_INVALID_INPUT):
                raise BitBox02Error(str(exc))
            raise exc
        except FirmwareVersionOutdatedException as exc:
            raise DeviceNotReadyError(str(exc))

    return cast(T, func)


# This class extends the HardwareWalletClient for BitBox02 specific things
class Bitbox02Client(HardwareWalletClient):
    def __init__(self, path: str, password: str = "", expert: bool = False) -> None:
        """
        Initializes a new BitBox02 client instance.
        """
        super().__init__(path, password=password, expert=expert)
        if password:
            raise BadArgumentError(
                "The BitBox02 does not accept a passphrase from the host. Please enable the passphrase option and enter the passphrase on the device during unlock."
            )

        hid_device = hid.device()
        hid_device.open_path(path.encode())
        self.transport = u2fhid.U2FHid(hid_device)
        self.device_path = path

        # use self.init() to access self.bb02.
        self.bb02: Optional[bitbox02.BitBox02] = None

        self.noise_config: BitBoxNoiseConfig = CLINoiseConfig()

    def set_noise_config(self, noise_config: BitBoxNoiseConfig) -> None:
        self.noise_config = noise_config

    def init(self, expect_initialized: Optional[bool] = True) -> bitbox02.BitBox02:
        if self.bb02 is not None:
            return self.bb02

        for device_info in devices.get_any_bitbox02s():
            if device_info["path"].decode() == self.device_path:
                bb02 = bitbox02.BitBox02(
                    transport=self.transport,
                    device_info=device_info,
                    noise_config=self.noise_config,
                )
                try:
                    bb02.check_min_version()
                except FirmwareVersionOutdatedException as exc:
                    sys.stderr.write("WARNING: {}\n".format(exc))
                    raise
            self.bb02 = bb02
            is_initialized = bb02.device_info()["initialized"]
            if expect_initialized is not None:
                if expect_initialized:
                    if not is_initialized:
                        raise HWWError(
                            "The BitBox02 must be initialized first.",
                            DEVICE_NOT_INITIALIZED,
                        )
                elif is_initialized:
                    raise UnavailableActionError(
                        "The BitBox02 must be wiped before setup."
                    )

            return bb02
        raise Exception(
            "Could not find the hid device info for path {}".format(self.device_path)
        )

    def close(self) -> None:
        self.transport.close()

    def get_master_fingerprint(self) -> bytes:
        """
        HWI by default retrieves the fingerprint at m/ by getting the xpub at m/0', which contains the parent fingerprint.
        The BitBox02 does not support querying arbitrary keypaths, but has an api call return the fingerprint at m/.
        """
        bb02 = self.init()
        return bb02.root_fingerprint()

    def prompt_pin(self) -> bool:
        raise UnavailableActionError(
            "The BitBox02 does not need a PIN sent from the host"
        )

    def send_pin(self, pin: str) -> bool:
        raise UnavailableActionError(
            "The BitBox02 does not need a PIN sent from the host"
        )

    def _get_coin(self) -> bitbox02.btc.BTCCoin:
        if self.chain != Chain.MAIN:
            return bitbox02.btc.TBTC
        return bitbox02.btc.BTC

    def _get_xpub(self, keypath: Sequence[int]) -> str:
        xpub_type = (
            bitbox02.btc.BTCPubRequest.TPUB
            if self.chain != Chain.MAIN
            else bitbox02.btc.BTCPubRequest.XPUB
        )
        return self.init().btc_xpub(
            keypath, coin=self._get_coin(), xpub_type=xpub_type, display=False
        )

    def get_pubkey_at_path(self, bip32_path: str) -> ExtendedKey:
        """
        Fetch the public key at the derivation path.

        The BitBox02 has strict keypath validation.

        The only accepted keypaths for xpubs are (as of firmware v9.4.0):

        - `m/49'/0'/<account'>` for `p2wpkh-p2sh` (segwit wrapped in P2SH)
        - `m/84'/0'/<account'>` for `p2wpkh` (native segwit v0)
        - `m/48'/0'/<account'>/2'` for p2wsh multisig (native segwit v0 multisig).
        - `m/48'/0'/<account'>/1'` for p2wsh-p2sh multisig (p2sh-wrapped segwit v0 multisig).
        - `m/48'/0'/<account'>` for p2wsh and p2wsh-p2sh multisig.

        `account'` can be between `0'` and `99'`.

        For address keypaths, append `/0/<address index>` for a receive and `/1/<change index>` for a change
        address. Up to `10000` addresses are supported.

        In testnet mode, the second element must be `1'` (e.g. `m/49'/1'/...`).

        Public keys for the Legacy address type (i.e. P2WPKH and P2SH multisig) derivation path is unsupported.
        """
        path_uint32s = parse_path(bip32_path)
        try:
            xpub_str = self._get_xpub(path_uint32s)
        except Bitbox02Exception as exc:
            raise BitBox02Error(str(exc))
        xpub = ExtendedKey.deserialize(xpub_str)
        return xpub

    def _maybe_register_script_config(
        self, script_config: bitbox02.btc.BTCScriptConfig, keypath: Sequence[int]
    ) -> None:
        bb02 = self.init()
        is_registered = bb02.btc_is_script_config_registered(
            self._get_coin(), script_config, keypath
        )
        if not is_registered:
            bb02.btc_register_script_config(
                coin=self._get_coin(),
                script_config=script_config,
                keypath=keypath,
                name="",  # enter name on the device
                xpub_type=bitbox02.btc.BTCRegisterScriptConfigRequest.AUTO_XPUB_TPUB,
            )

    def _multisig_scriptconfig(
        self,
        threshold: int,
        origin_infos: Mapping[bytes, KeyOriginInfo],
        script_type: bitbox02.btc.BTCScriptConfig.Multisig.ScriptType,
    ) -> Tuple[bytes, bitbox02.btc.BTCScriptConfigWithKeypath]:
        """
        From a threshold, {xpub: KeyOriginInfo} mapping and multisig script type,
        return our xpub and the BitBox02 multisig script config.
        """
        # Figure out which of the cosigners is us.
        device_fingerprint = self.get_master_fingerprint()
        our_xpub_index = None
        our_account_keypath = None

        xpubs: List[bytes] = []
        for i, (xpub, keyinfo) in builtins.enumerate(origin_infos.items()):
            xpubs.append(xpub)
            if device_fingerprint == keyinfo.fingerprint and keyinfo.path:
                if _xpubs_equal_ignoring_version(
                    base58.b58decode_check(self._get_xpub(keyinfo.path)), xpub
                ):
                    our_xpub_index = i
                    our_account_keypath = keyinfo.path

        if our_xpub_index is None:
            raise BadArgumentError("This BitBox02 is not one of the cosigners")
        assert our_account_keypath

        if len(xpubs) != len(set(xpubs)):
            raise BadArgumentError("Duplicate xpubs not supported")

        return (
            xpubs[our_xpub_index],
            bitbox02.btc.BTCScriptConfigWithKeypath(
                script_config=bitbox02.btc.BTCScriptConfig(
                    multisig=bitbox02.btc.BTCScriptConfig.Multisig(
                        threshold=threshold,
                        xpubs=[util.parse_xpub(base58.b58encode_check(xpub).decode()) for xpub in xpubs],
                        our_xpub_index=our_xpub_index,
                        script_type=script_type,
                    )
                ),
                keypath=our_account_keypath,
            ),
        )

    @bitbox02_exception
    def display_singlesig_address(
        self,
        bip32_path: str,
        addr_type: AddressType,
    ) -> str:
        if addr_type == AddressType.SH_WIT:
            script_config = bitbox02.btc.BTCScriptConfig(
                simple_type=bitbox02.btc.BTCScriptConfig.P2WPKH_P2SH
            )
        elif addr_type == AddressType.WIT:
            script_config = bitbox02.btc.BTCScriptConfig(
                simple_type=bitbox02.btc.BTCScriptConfig.P2WPKH
            )
        elif addr_type == AddressType.LEGACY:
            raise UnavailableActionError(
                "The BitBox02 does not support legacy p2pkh addresses"
            )
        else:
            raise BadArgumentError("Unknown address type")
        address = self.init().btc_address(
            parse_path(bip32_path),
            coin=self._get_coin(),
            script_config=script_config,
            display=True,
        )
        return address

    @bitbox02_exception
    def display_multisig_address(
        self,
        addr_type: AddressType,
        multisig: MultisigDescriptor,
    ) -> str:
        if not multisig.is_sorted:
            raise BadArgumentError("BitBox02 only supports sortedmulti descriptors")

        path_suffixes = set(p.deriv_path for p in multisig.pubkeys)
        if len(path_suffixes) != 1:
            # Path suffix refers to the path after the account-level xpub, usually /<change>/<address>.
            # The BitBox02 currently enforces that all of them are the same.
            raise BadArgumentError("All multisig path suffixes must be the same")

        # Figure out which of the cosigners is us.
        key_origin_infos = {}
        keypaths = {}
        for pk in multisig.pubkeys:
            assert pk.extkey and pk.origin
            key_origin_infos[pk.extkey.serialize()] = pk.origin
            keypaths[pk.extkey.serialize()] = pk.get_full_derivation_path(0)

        if addr_type == AddressType.SH_WIT:
            script_type = bitbox02.btc.BTCScriptConfig.Multisig.P2WSH_P2SH
        elif addr_type == AddressType.WIT:
            script_type = bitbox02.btc.BTCScriptConfig.Multisig.P2WSH
        else:
            raise BadArgumentError(
                "BitBox02 currently only supports the following multisig script types: P2WSH, P2WSH_P2SH"
            )
        our_xpub, script_config_with_keypath = self._multisig_scriptconfig(
            multisig.thresh, key_origin_infos, script_type
        )
        script_config = script_config_with_keypath.script_config
        account_keypath: Sequence[int] = script_config_with_keypath.keypath
        self._maybe_register_script_config(script_config, account_keypath)
        keypath = parse_path(keypaths[our_xpub])

        bb02 = self.init()
        address = bb02.btc_address(
            keypath, coin=self._get_coin(), script_config=script_config, display=True
        )
        return address

    @bitbox02_exception
    def sign_tx(self, psbt: PSBT) -> PSBT:
        """
        Sign a transaction with the BitBox02.

        he BitBox02 allows mixing inputs of different script types (e.g. and `p2wpkh-p2sh` `p2wpkh`), as
        long as the keypaths use the appropriate bip44 purpose field per input (e.g. `49'` and `84'`) and
        all account indexes are the same.

        Transactions with legacy inputs are not supported.
        """
        def find_our_key(
            keypaths: Dict[bytes, KeyOriginInfo]
        ) -> Tuple[Optional[bytes], Optional[Sequence[int]]]:
            """
            Keypaths is a map of pubkey to hd keypath, where the first element in the keypath is the master fingerprint. We attempt to find the key which belongs to the BitBox02 by matching the fingerprint, and then matching the pubkey.
            Returns the pubkey and the keypath, without the fingerprint.
            """
            for pubkey, origin in keypaths.items():
                # Cheap check if the key is ours.
                if origin.fingerprint != master_fp:
                    continue

                # Expensive check if the key is ours.
                # TODO: check for fingerprint collision
                # keypath_account = keypath[:-2]

                return pubkey, origin.path
            return None, None

        script_configs: List[bitbox02.btc.BTCScriptConfigWithKeypath] = []

        def add_script_config(
            script_config: bitbox02.btc.BTCScriptConfigWithKeypath
        ) -> int:
            # Find index of script config if already added.
            script_config_index = next(
                (
                    i
                    for i, e in builtins.enumerate(script_configs)
                    if e.SerializeToString() == script_config.SerializeToString()
                ),
                None,
            )
            if script_config_index is not None:
                return script_config_index
            script_configs.append(script_config)
            return len(script_configs) - 1

        def script_config_from_utxo(
            output: CTxOut,
            keypath: Sequence[int],
            redeem_script: bytes,
            witness_script: bytes,
        ) -> bitbox02.btc.BTCScriptConfigWithKeypath:
            if is_p2pkh(output.scriptPubKey):
                raise BadArgumentError(
                    "The BitBox02 does not support legacy p2pkh scripts"
                )
            if is_p2wpkh(output.scriptPubKey):
                return bitbox02.btc.BTCScriptConfigWithKeypath(
                    script_config=bitbox02.btc.BTCScriptConfig(
                        simple_type=bitbox02.btc.BTCScriptConfig.P2WPKH
                    ),
                    keypath=_keypath_hardened_prefix(keypath),
                )
            if output.is_p2sh() and is_p2wpkh(redeem_script):
                return bitbox02.btc.BTCScriptConfigWithKeypath(
                    script_config=bitbox02.btc.BTCScriptConfig(
                        simple_type=bitbox02.btc.BTCScriptConfig.P2WPKH_P2SH
                    ),
                    keypath=_keypath_hardened_prefix(keypath),
                )
            # Check for segwit multisig (p2wsh or p2wsh-p2sh).
            is_p2wsh_p2sh = output.is_p2sh() and is_p2wsh(redeem_script)
            if output.is_p2wsh() or is_p2wsh_p2sh:
                multisig = parse_multisig(witness_script)
                if multisig:
                    threshold, _ = multisig
                    # We assume that all xpubs in the PSBT are part of the multisig. This is okay
                    # since the BitBox02 enforces the same script type for all inputs and
                    # changes. If that should change, we need to find and use the subset of xpubs
                    # corresponding to the public keys in the current multisig script.
                    _, script_config = self._multisig_scriptconfig(
                        threshold,
                        psbt.xpub,
                        bitbox02.btc.BTCScriptConfig.Multisig.P2WSH
                        if output.is_p2wsh()
                        else bitbox02.btc.BTCScriptConfig.Multisig.P2WSH_P2SH,
                    )
                    return script_config

            raise BadArgumentError("Input or change script type not recognized.")

        master_fp = self.get_master_fingerprint()

        inputs: List[bitbox02.BTCInputType] = []

        bip44_account = None

        # One pubkey per input. The pubkey identifies the key per input with which we sign. There
        # must be exactly one pubkey per input that belongs to the BitBox02.
        found_pubkeys: List[bytes] = []

        for input_index, (psbt_in, tx_in) in builtins.enumerate(
            zip(psbt.inputs, psbt.tx.vin)
        ):
            if psbt_in.sighash and psbt_in.sighash != 1:
                raise BadArgumentError(
                    "The BitBox02 only supports SIGHASH_ALL. Found sighash: {}".format(
                        psbt_in.sighash
                    )
                )

            utxo = None
            prevtx = None

            # psbt_in.witness_utxo was originally used for segwit utxo's, but since it was
            # discovered that the amounts are not correctly committed to in the segwit sighash, the
            # full prevtx (non_witness_utxo) is supplied for both segwit and non-segwit inputs.
            # See
            # - https://medium.com/shiftcrypto/bitbox-app-firmware-update-6-2020-c70f733a5330
            # - https://blog.trezor.io/details-of-firmware-updates-for-trezor-one-version-1-9-1-and-trezor-model-t-version-2-3-1-1eba8f60f2dd.
            # - https://github.com/zkSNACKs/WalletWasabi/pull/3822
            # The BitBox02 for now requires the prevtx, at least until Taproot activates.

            if psbt_in.non_witness_utxo:
                if tx_in.prevout.hash != psbt_in.non_witness_utxo.sha256:
                    raise BadArgumentError(
                        "Input {} has a non_witness_utxo with the wrong hash".format(
                            input_index
                        )
                    )
                utxo = psbt_in.non_witness_utxo.vout[tx_in.prevout.n]
                prevtx = psbt_in.non_witness_utxo
            elif psbt_in.witness_utxo:
                utxo = psbt_in.witness_utxo
            if utxo is None:
                raise BadArgumentError("No utxo found for input {}".format(input_index))
            if prevtx is None:
                raise BadArgumentError(
                    "Previous transaction missing for input {}".format(input_index)
                )

            found_pubkey, keypath = find_our_key(psbt_in.hd_keypaths)
            if not found_pubkey:
                raise BadArgumentError("No key found for input {}".format(input_index))
            assert keypath is not None
            found_pubkeys.append(found_pubkey)

            if bip44_account is None:
                bip44_account = keypath[2]
            elif bip44_account != keypath[2]:
                raise BadArgumentError(
                    "The bip44 account index must be the same for all inputs and changes"
                )

            script_config_index = add_script_config(
                script_config_from_utxo(
                    utxo, keypath, psbt_in.redeem_script, psbt_in.witness_script
                )
            )
            inputs.append(
                {
                    "prev_out_hash": ser_uint256(tx_in.prevout.hash),
                    "prev_out_index": tx_in.prevout.n,
                    "prev_out_value": utxo.nValue,
                    "sequence": tx_in.nSequence,
                    "keypath": keypath,
                    "script_config_index": script_config_index,
                    "prev_tx": {
                        "version": prevtx.nVersion,
                        "locktime": prevtx.nLockTime,
                        "inputs": [
                            {
                                "prev_out_hash": ser_uint256(prev_in.prevout.hash),
                                "prev_out_index": prev_in.prevout.n,
                                "signature_script": prev_in.scriptSig,
                                "sequence": prev_in.nSequence,
                            }
                            for prev_in in prevtx.vin
                        ],
                        "outputs": [
                            {
                                "value": prev_out.nValue,
                                "pubkey_script": prev_out.scriptPubKey,
                            }
                            for prev_out in prevtx.vout
                        ],
                    },
                }
            )

        outputs: List[bitbox02.BTCOutputType] = []
        for output_index, (psbt_out, tx_out) in builtins.enumerate(
            zip(psbt.outputs, psbt.tx.vout)
        ):
            _, keypath = find_our_key(psbt_out.hd_keypaths)
            is_change = keypath and keypath[-2] == 1
            if is_change:
                assert keypath is not None
                script_config_index = add_script_config(
                    script_config_from_utxo(
                        tx_out, keypath, psbt_out.redeem_script, psbt_out.witness_script
                    )
                )
                outputs.append(
                    bitbox02.BTCOutputInternal(
                        keypath=keypath,
                        value=tx_out.nValue,
                        script_config_index=script_config_index,
                    )
                )
            else:
                if tx_out.is_p2pkh():
                    output_type = bitbox02.btc.P2PKH
                    output_hash = tx_out.scriptPubKey[3:23]
                elif is_p2wpkh(tx_out.scriptPubKey):
                    output_type = bitbox02.btc.P2WPKH
                    output_hash = tx_out.scriptPubKey[2:]
                elif tx_out.is_p2sh():
                    output_type = bitbox02.btc.P2SH
                    output_hash = tx_out.scriptPubKey[2:22]
                elif is_p2wsh(tx_out.scriptPubKey):
                    output_type = bitbox02.btc.P2WSH
                    output_hash = tx_out.scriptPubKey[2:]
                else:
                    raise BadArgumentError(
                        "Output type not recognized of output {}".format(output_index)
                    )

                outputs.append(
                    bitbox02.BTCOutputExternal(
                        output_type=output_type,
                        output_hash=output_hash,
                        value=tx_out.nValue,
                    )
                )

        assert bip44_account is not None
        if (
            len(script_configs) == 1
            and script_configs[0].script_config.WhichOneof("config") == "multisig"
        ):
            self._maybe_register_script_config(
                script_configs[0].script_config, script_configs[0].keypath
            )

        sigs = self.init().btc_sign(
            self._get_coin(),
            script_configs,
            inputs=inputs,
            outputs=outputs,
            locktime=psbt.tx.nLockTime,
            version=psbt.tx.nVersion,
        )

        for (_, sig), pubkey, psbt_in in zip(sigs, found_pubkeys, psbt.inputs):
            r, s = sig[:32], sig[32:64]
            # ser_sig_der() adds SIGHASH_ALL
            psbt_in.partial_sigs[pubkey] = ser_sig_der(r, s)

        return psbt

    def sign_message(
        self, message: Union[str, bytes], bip32_path: str
    ) -> str:
        raise UnavailableActionError("The BitBox02 does not support 'signmessage'")

    @bitbox02_exception
    def toggle_passphrase(self) -> bool:
        bb02 = self.init()
        info = bb02.device_info()
        if info["mnemonic_passphrase_enabled"]:
            bb02.disable_mnemonic_passphrase()
        else:
            bb02.enable_mnemonic_passphrase()
        return True

    @bitbox02_exception
    def setup_device(
        self, label: str = "", passphrase: str = ""
    ) -> bool:
        if passphrase:
            raise UnavailableActionError(
                "Passphrase not needed when setting up a BitBox02."
            )

        bb02 = self.init(expect_initialized=False)

        if label:
            bb02.set_device_name(label)
        if not bb02.set_password():
            return False
        return bb02.create_backup()

    @bitbox02_exception
    def wipe_device(self) -> bool:
        return self.init().reset()

    @bitbox02_exception
    def backup_device(
        self, label: str = "", passphrase: str = ""
    ) -> bool:
        if label or passphrase:
            raise UnavailableActionError(
                "Label/passphrase not needed when exporting mnemonic from the BitBox02."
            )

        self.init().show_mnemonic()
        return True

    @bitbox02_exception
    def restore_device(
        self, label: str = "", word_count: int = 24
    ) -> bool:
        bb02 = self.init(expect_initialized=False)

        if label:
            bb02.set_device_name(label)

        bb02.restore_from_mnemonic()
        return True



# Automatically generated by pb2py
# fmt: off
if __debug__:
    try:
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass

Strict: Literal[0] = 0
PromptAlways: Literal[1] = 1
PromptTemporarily: Literal[2] = 2



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'getkeypooloptionsdialog.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_GetKeypoolOptionsDialog(object):
    def setupUi(self, GetKeypoolOptionsDialog):
        if not GetKeypoolOptionsDialog.objectName():
            GetKeypoolOptionsDialog.setObjectName(u"GetKeypoolOptionsDialog")
        GetKeypoolOptionsDialog.resize(440, 224)
        self.buttonBox = QDialogButtonBox(GetKeypoolOptionsDialog)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setGeometry(QRect(80, 180, 341, 32))
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok)
        self.label = QLabel(GetKeypoolOptionsDialog)
        self.label.setObjectName(u"label")
        self.label.setGeometry(QRect(20, 20, 41, 18))
        self.label_2 = QLabel(GetKeypoolOptionsDialog)
        self.label_2.setObjectName(u"label_2")
        self.label_2.setGeometry(QRect(20, 60, 31, 18))
        self.start_spinbox = QSpinBox(GetKeypoolOptionsDialog)
        self.start_spinbox.setObjectName(u"start_spinbox")
        self.start_spinbox.setGeometry(QRect(80, 10, 161, 32))
        self.start_spinbox.setMaximum(2147483647)
        self.end_spinbox = QSpinBox(GetKeypoolOptionsDialog)
        self.end_spinbox.setObjectName(u"end_spinbox")
        self.end_spinbox.setGeometry(QRect(80, 50, 161, 32))
        self.end_spinbox.setMaximum(2147483647)
        self.end_spinbox.setValue(1000)
        self.internal_checkbox = QCheckBox(GetKeypoolOptionsDialog)
        self.internal_checkbox.setObjectName(u"internal_checkbox")
        self.internal_checkbox.setGeometry(QRect(280, 10, 88, 22))
        self.keypool_checkbox = QCheckBox(GetKeypoolOptionsDialog)
        self.keypool_checkbox.setObjectName(u"keypool_checkbox")
        self.keypool_checkbox.setGeometry(QRect(280, 40, 88, 22))
        self.keypool_checkbox.setChecked(True)
        self.groupBox = QGroupBox(GetKeypoolOptionsDialog)
        self.groupBox.setObjectName(u"groupBox")
        self.groupBox.setGeometry(QRect(280, 70, 141, 101))
        self.sh_wpkh_radio = QRadioButton(self.groupBox)
        self.sh_wpkh_radio.setObjectName(u"sh_wpkh_radio")
        self.sh_wpkh_radio.setGeometry(QRect(10, 10, 121, 22))
        self.sh_wpkh_radio.setChecked(True)
        self.wpkh_radio = QRadioButton(self.groupBox)
        self.wpkh_radio.setObjectName(u"wpkh_radio")
        self.wpkh_radio.setGeometry(QRect(10, 40, 105, 22))
        self.pkh_radio = QRadioButton(self.groupBox)
        self.pkh_radio.setObjectName(u"pkh_radio")
        self.pkh_radio.setGeometry(QRect(10, 70, 105, 22))
        self.groupBox_2 = QGroupBox(GetKeypoolOptionsDialog)
        self.groupBox_2.setObjectName(u"groupBox_2")
        self.groupBox_2.setGeometry(QRect(10, 90, 231, 91))
        self.account_spinbox = QSpinBox(self.groupBox_2)
        self.account_spinbox.setObjectName(u"account_spinbox")
        self.account_spinbox.setGeometry(QRect(100, 10, 111, 32))
        self.account_spinbox.setMaximum(2147483647)
        self.account_spinbox.setValue(0)
        self.account_radio = QRadioButton(self.groupBox_2)
        self.account_radio.setObjectName(u"account_radio")
        self.account_radio.setGeometry(QRect(10, 10, 81, 22))
        self.account_radio.setChecked(True)
        self.path_radio = QRadioButton(self.groupBox_2)
        self.path_radio.setObjectName(u"path_radio")
        self.path_radio.setGeometry(QRect(10, 50, 61, 22))
        self.path_lineedit = QLineEdit(self.groupBox_2)
        self.path_lineedit.setObjectName(u"path_lineedit")
        self.path_lineedit.setEnabled(False)
        self.path_lineedit.setGeometry(QRect(80, 50, 141, 32))

        self.retranslateUi(GetKeypoolOptionsDialog)
        self.buttonBox.accepted.connect(GetKeypoolOptionsDialog.accept)
        self.buttonBox.rejected.connect(GetKeypoolOptionsDialog.reject)

        QMetaObject.connectSlotsByName(GetKeypoolOptionsDialog)
    # setupUi

    def retranslateUi(self, GetKeypoolOptionsDialog):
        GetKeypoolOptionsDialog.setWindowTitle(QCoreApplication.translate("GetKeypoolOptionsDialog", u"Dialog", None))
        self.label.setText(QCoreApplication.translate("GetKeypoolOptionsDialog", u"Start", None))
        self.label_2.setText(QCoreApplication.translate("GetKeypoolOptionsDialog", u"End", None))
        self.internal_checkbox.setText(QCoreApplication.translate("GetKeypoolOptionsDialog", u"Internal", None))
        self.keypool_checkbox.setText(QCoreApplication.translate("GetKeypoolOptionsDialog", u"keypool", None))
        self.groupBox.setTitle("")
        self.sh_wpkh_radio.setText(QCoreApplication.translate("GetKeypoolOptionsDialog", u"P2SH-P2WPKH", None))
        self.wpkh_radio.setText(QCoreApplication.translate("GetKeypoolOptionsDialog", u"P2WPKH", None))
        self.pkh_radio.setText(QCoreApplication.translate("GetKeypoolOptionsDialog", u"P2PKH", None))
        self.groupBox_2.setTitle("")
        self.account_radio.setText(QCoreApplication.translate("GetKeypoolOptionsDialog", u"Account", None))
        self.path_radio.setText(QCoreApplication.translate("GetKeypoolOptionsDialog", u"Path", None))
        self.path_lineedit.setText(QCoreApplication.translate("GetKeypoolOptionsDialog", u"m/0'/0'/*", None))
    # retranslateUi




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'displayaddressdialog.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_DisplayAddressDialog(object):
    def setupUi(self, DisplayAddressDialog):
        if not DisplayAddressDialog.objectName():
            DisplayAddressDialog.setObjectName(u"DisplayAddressDialog")
        DisplayAddressDialog.resize(469, 196)
        self.buttonBox = QDialogButtonBox(DisplayAddressDialog)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setGeometry(QRect(350, 150, 101, 32))
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Close)
        self.path_lineedit = QLineEdit(DisplayAddressDialog)
        self.path_lineedit.setObjectName(u"path_lineedit")
        self.path_lineedit.setGeometry(QRect(120, 10, 331, 32))
        self.label = QLabel(DisplayAddressDialog)
        self.label.setObjectName(u"label")
        self.label.setGeometry(QRect(10, 20, 111, 18))
        self.go_button = QPushButton(DisplayAddressDialog)
        self.go_button.setObjectName(u"go_button")
        self.go_button.setGeometry(QRect(410, 50, 41, 41))
        self.go_button.setAutoDefault(False)
        self.label_2 = QLabel(DisplayAddressDialog)
        self.label_2.setObjectName(u"label_2")
        self.label_2.setGeometry(QRect(10, 120, 58, 18))
        self.address_lineedit = QLineEdit(DisplayAddressDialog)
        self.address_lineedit.setObjectName(u"address_lineedit")
        self.address_lineedit.setGeometry(QRect(70, 110, 381, 32))
        self.address_lineedit.setReadOnly(True)
        self.type_groupbox = QGroupBox(DisplayAddressDialog)
        self.type_groupbox.setObjectName(u"type_groupbox")
        self.type_groupbox.setGeometry(QRect(10, 50, 381, 40))
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.type_groupbox.sizePolicy().hasHeightForWidth())
        self.type_groupbox.setSizePolicy(sizePolicy)
        self.type_groupbox.setMinimumSize(QSize(0, 30))
        self.sh_wpkh_radio = QRadioButton(self.type_groupbox)
        self.sh_wpkh_radio.setObjectName(u"sh_wpkh_radio")
        self.sh_wpkh_radio.setGeometry(QRect(10, 10, 121, 22))
        self.sh_wpkh_radio.setChecked(True)
        self.wpkh_radio = QRadioButton(self.type_groupbox)
        self.wpkh_radio.setObjectName(u"wpkh_radio")
        self.wpkh_radio.setGeometry(QRect(150, 10, 91, 22))
        self.radioButton = QRadioButton(self.type_groupbox)
        self.radioButton.setObjectName(u"radioButton")
        self.radioButton.setGeometry(QRect(260, 10, 105, 22))

        self.retranslateUi(DisplayAddressDialog)
        self.buttonBox.accepted.connect(DisplayAddressDialog.accept)
        self.buttonBox.rejected.connect(DisplayAddressDialog.reject)

        self.go_button.setDefault(True)


        QMetaObject.connectSlotsByName(DisplayAddressDialog)
    # setupUi

    def retranslateUi(self, DisplayAddressDialog):
        DisplayAddressDialog.setWindowTitle(QCoreApplication.translate("DisplayAddressDialog", u"Dialog", None))
        self.label.setText(QCoreApplication.translate("DisplayAddressDialog", u"Derivation Path", None))
        self.go_button.setText(QCoreApplication.translate("DisplayAddressDialog", u"Go", None))
        self.label_2.setText(QCoreApplication.translate("DisplayAddressDialog", u"Address", None))
        self.type_groupbox.setTitle("")
        self.sh_wpkh_radio.setText(QCoreApplication.translate("DisplayAddressDialog", u"P2SH-P2WPKH", None))
        self.wpkh_radio.setText(QCoreApplication.translate("DisplayAddressDialog", u"P2WPKH", None))
        self.radioButton.setText(QCoreApplication.translate("DisplayAddressDialog", u"P2PKH", None))
    # retranslateUi




# Automatically generated by pb2py
# fmt: off
if __debug__:
    try:
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass

Initialize: Literal[0] = 0
Ping: Literal[1] = 1
Success: Literal[2] = 2
Failure: Literal[3] = 3
ChangePin: Literal[4] = 4
WipeDevice: Literal[5] = 5
GetEntropy: Literal[9] = 9
Entropy: Literal[10] = 10
LoadDevice: Literal[13] = 13
ResetDevice: Literal[14] = 14
Features: Literal[17] = 17
PinMatrixRequest: Literal[18] = 18
PinMatrixAck: Literal[19] = 19
Cancel: Literal[20] = 20
LockDevice: Literal[24] = 24
ApplySettings: Literal[25] = 25
ButtonRequest: Literal[26] = 26
ButtonAck: Literal[27] = 27
ApplyFlags: Literal[28] = 28
BackupDevice: Literal[34] = 34
EntropyRequest: Literal[35] = 35
EntropyAck: Literal[36] = 36
PassphraseRequest: Literal[41] = 41
PassphraseAck: Literal[42] = 42
RecoveryDevice: Literal[45] = 45
WordRequest: Literal[46] = 46
WordAck: Literal[47] = 47
GetFeatures: Literal[55] = 55
EndSession: Literal[83] = 83
DoPreauthorized: Literal[84] = 84
PreauthorizedRequest: Literal[85] = 85
Deprecated_PassphraseStateRequest: Literal[77] = 77
Deprecated_PassphraseStateAck: Literal[78] = 78
FirmwareErase: Literal[6] = 6
FirmwareUpload: Literal[7] = 7
FirmwareRequest: Literal[8] = 8
SelfTest: Literal[32] = 32
GetPublicKey: Literal[11] = 11
PublicKey: Literal[12] = 12
SignTx: Literal[15] = 15
TxRequest: Literal[21] = 21
TxAck: Literal[22] = 22
GetAddress: Literal[29] = 29
Address: Literal[30] = 30
SignMessage: Literal[38] = 38
VerifyMessage: Literal[39] = 39
MessageSignature: Literal[40] = 40
GetOwnershipId: Literal[43] = 43
OwnershipId: Literal[44] = 44
GetOwnershipProof: Literal[49] = 49
OwnershipProof: Literal[50] = 50
AuthorizeCoinJoin: Literal[51] = 51
DebugLinkDecision: Literal[100] = 100
DebugLinkGetState: Literal[101] = 101
DebugLinkState: Literal[102] = 102
DebugLinkStop: Literal[103] = 103
DebugLinkMemoryRead: Literal[110] = 110
DebugLinkMemory: Literal[111] = 111
DebugLinkMemoryWrite: Literal[112] = 112
DebugLinkFlashErase: Literal[113] = 113
DebugLinkReseedRandom: Literal[9002] = 9002
DebugLinkRecordScreen: Literal[9003] = 9003
DebugLinkShowText: Literal[9004] = 9004
DebugLinkEraseSdCard: Literal[9005] = 9005
DebugLinkWatchLayout: Literal[9006] = 9006



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'bitbox02pairing.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_BitBox02PairingDialog(object):
    def setupUi(self, BitBox02PairingDialog):
        if not BitBox02PairingDialog.objectName():
            BitBox02PairingDialog.setObjectName(u"BitBox02PairingDialog")
        BitBox02PairingDialog.setWindowModality(Qt.WindowModal)
        BitBox02PairingDialog.resize(400, 209)
        self.buttonBox = QDialogButtonBox(BitBox02PairingDialog)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setGeometry(QRect(30, 160, 341, 32))
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.No|QDialogButtonBox.Yes)
        self.pairingCode = QLabel(BitBox02PairingDialog)
        self.pairingCode.setObjectName(u"pairingCode")
        self.pairingCode.setGeometry(QRect(20, 80, 331, 61))
        font = QFont()
        font.setFamily(u"DejaVu Sans Mono")
        font.setPointSize(15)
        font.setBold(True)
        font.setWeight(75)
        self.pairingCode.setFont(font)
        self.pairingCode.setTextFormat(Qt.RichText)
        self.pairingCode.setAlignment(Qt.AlignCenter)
        self.label_2 = QLabel(BitBox02PairingDialog)
        self.label_2.setObjectName(u"label_2")
        self.label_2.setGeometry(QRect(20, 10, 351, 61))
        font1 = QFont()
        font1.setPointSize(11)
        self.label_2.setFont(font1)
        self.label_2.setTextFormat(Qt.PlainText)

        self.retranslateUi(BitBox02PairingDialog)
        self.buttonBox.accepted.connect(BitBox02PairingDialog.accept)
        self.buttonBox.rejected.connect(BitBox02PairingDialog.reject)

        QMetaObject.connectSlotsByName(BitBox02PairingDialog)
    # setupUi

    def retranslateUi(self, BitBox02PairingDialog):
        BitBox02PairingDialog.setWindowTitle(QCoreApplication.translate("BitBox02PairingDialog", u"Dialog", None))
        self.pairingCode.setText("")
        self.label_2.setText(QCoreApplication.translate("BitBox02PairingDialog", u"Please verify the pairing code matches what is\n"
"shown on your BitBox02.", None))
    # retranslateUi




# Autogen'ed file, don't edit. See bootloader/mk-sigheader.h for original


# Our simple firmware header.
# Although called a header, this data is placed into the middle of the binary.
# It is located at start of firmware + 16k - sizeof(heaer). This is a gap unused in normal
# micropython layout. Exactly the last 64 bytes (signature) should be left out of
# the checksum. We do checksum areas beyond the end of the last byte of firmware (up to length)
# and expect those regions to be unprogrammed flash (ones).
# - timestamp must increase with each upgrade (downgrade protection)
# - version_string is for humans only
# - pubkey_num indicates which pubkey was used for signature
# - firmware_length, must be:
# - bigger than minimum length, less than max
# - 512-byte aligned
# - bootloader assumes the flash filesystem (FAT FS) follows the firmware.
# - this C header file is somewhat parsed and used by python signature-adding code
# - timestamp is YYMMDDHHMMSS0000 in BCD


FW_HEADER_SIZE = 128
FW_HEADER_OFFSET = (0x4000-FW_HEADER_SIZE)

FW_HEADER_MAGIC = 0xCC001234

# arbitrary min size
FW_MIN_LENGTH = (256*1024)
# absolute max: 1MB flash - 32k for bootloader
# practical limit for our-protocol USB upgrades: 786432 (or else settings damaged)
FW_MAX_LENGTH = (0x100000 - 0x8000)

# Arguments to be used w/ python's struct module.
FWH_PY_FORMAT = "<I8s8sII36s64s"
FWH_PY_VALUES = "magic_value timestamp version_string pubkey_num firmware_length future signature"
FWH_NUM_FUTURE = 9
FWH_PK_NUM_OFFSET = 20

# There is a copy of the header at this location in RAM, copied by bootloader
# **after** it has been verified. Cannot write to this area, or you will be reset!
RAM_HEADER_BASE = 0x10007c20

# Original copy of header, as recorded in flash/firmware file.
FLASH_HEADER_BASE = 0x0800bf80

# EOF



# Automatically generated by pb2py
# fmt: off
if __debug__:
    try:
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass

Current: Literal[1] = 1
NewFirst: Literal[2] = 2
NewSecond: Literal[3] = 3
WipeCodeFirst: Literal[4] = 4
WipeCodeSecond: Literal[5] = 5



# Automatically generated by pb2py
# fmt: off
if __debug__:
    try:
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass

NORMAL: Literal[0] = 0
BOLD: Literal[1] = 1
MONO: Literal[2] = 2
BR: Literal[4] = 4
BR_HALF: Literal[5] = 5
SET_COLOR: Literal[6] = 6



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
        EnumTypePinMatrixRequestType = Literal[1, 2, 3, 4, 5]
    except ImportError:
        pass


class PinMatrixRequest(p.MessageType):
    MESSAGE_WIRE_TYPE = 18

    def __init__(
        self,
        *,
        type: EnumTypePinMatrixRequestType = None,
    ) -> None:
        self.type = type

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('type', p.EnumType("PinMatrixRequestType", (1, 2, 3, 4, 5)), None),
        }



# Automatically generated by pb2py
# fmt: off
if __debug__:
    try:
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass

Other: Literal[1] = 1
FeeOverThreshold: Literal[2] = 2
ConfirmOutput: Literal[3] = 3
ResetDevice: Literal[4] = 4
ConfirmWord: Literal[5] = 5
WipeDevice: Literal[6] = 6
ProtectCall: Literal[7] = 7
SignTx: Literal[8] = 8
FirmwareCheck: Literal[9] = 9
Address: Literal[10] = 10
PublicKey: Literal[11] = 11
MnemonicWordCount: Literal[12] = 12
MnemonicInput: Literal[13] = 13
_Deprecated_ButtonRequest_PassphraseType: Literal[14] = 14
UnknownDerivationPath: Literal[15] = 15
RecoveryHomepage: Literal[16] = 16
Success: Literal[17] = 17
Warning: Literal[18] = 18
PassphraseEntry: Literal[19] = 19
PinEntry: Literal[20] = 20



# Automatically generated by pb2py
# fmt: off
if __debug__:
    try:
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass

Bip39: Literal[0] = 0
Slip39_Basic: Literal[1] = 1
Slip39_Advanced: Literal[2] = 2



#
# Details of our USB level protocol. Shared file between desktop and embedded.
#
# - first 4 bytes of all messages is the command code or response code
# - use <I and <H, never >H
#
from struct import pack, unpack_from
from .constants import *

class CCProtoError(RuntimeError):
    def __str__(self):
        return self.args[0]

class CCFramingError(CCProtoError):
    # Typically framing errors are caused by multiple
    # programs trying to talk to Coldcard at same time,
    # and the encryption state gets confused.
    pass

class CCUserRefused(RuntimeError):
    def __str__(self):
        return 'You refused permission to do the operation'

class CCBusyError(RuntimeError):
    def __str__(self):
        return 'Coldcard is handling another request right now'

class CCProtocolPacker:
    # returns a lamba that will take correct args
    # and then give you a binary string to encode the
    # request

    @staticmethod
    def logout():
        return pack('4s', b'logo')

    @staticmethod
    def reboot():
        return pack('4s', b'rebo')

    @staticmethod
    def version():
        # returns a string, with newline separators
        return pack('4s', b'vers')

    @staticmethod
    def ping(msg):
        # returns whatever binary you give it
        return b'ping' + bytes(msg)

    @staticmethod
    def bip39_passphrase(pw):
        return b'pass' + bytes(pw, 'utf8')

    @staticmethod
    def get_passphrase_done():
        # poll completion of BIP39 encryption change (provides root xpub)
        return b'pwok'

    @staticmethod
    def check_mitm():
        return b'mitm'

    @staticmethod
    def start_backup():
        # prompts user with password for encrytped backup
        return b'back'

    @staticmethod
    def encrypt_start(device_pubkey, version=0x1):
        assert len(device_pubkey) == 64, "want uncompressed 64-byte pubkey, no prefix byte"
        return pack('<4sI64s', b'ncry', version, device_pubkey)

    @staticmethod
    def upload(offset, total_size, data):
        # note: see MAX_MSG_LEN above
        assert len(data) <= MAX_MSG_LEN, 'badlen'
        return pack('<4sII', b'upld', offset, total_size) + data

    @staticmethod
    def download(offset, length, file_number=0):
        assert 0 <= file_number < 2
        return pack('<4sIII', b'dwld', offset, length, file_number)

    @staticmethod
    def sha256():
        return b'sha2'

    @staticmethod
    def sign_transaction(length, file_sha, finalize=False, flags=0x0):
        # must have already uploaded binary, and give expected sha256
        assert len(file_sha) == 32
        flags |= (STXN_FINALIZE if finalize else 0x00)
        return pack('<4sII32s', b'stxn', length, int(flags), file_sha)

    @staticmethod
    def sign_message(raw_msg, subpath='m', addr_fmt=AF_CLASSIC):
        # only begins user interaction
        return pack('<4sIII', b'smsg', addr_fmt, len(subpath), len(raw_msg)) \
                    + subpath.encode('ascii') + raw_msg

    @staticmethod
    def get_signed_msg():
        # poll completion/results of message signing
        return b'smok'

    @staticmethod
    def get_backup_file():
        # poll completion/results of backup
        return b'bkok'

    @staticmethod
    def get_signed_txn():
        # poll completion/results of transaction signing
        return b'stok'

    @staticmethod
    def multisig_enroll(length, file_sha):
        # multisig details must already be uploaded as a text file, this starts approval process.
        assert len(file_sha) == 32
        return pack('<4sI32s', b'enrl', length, file_sha)

    @staticmethod
    def multisig_check(M, N, xfp_xor):
        # do we have a wallet already that matches M+N and xor(*xfps)?
        return pack('<4s3I', b'msck', M, N, xfp_xor)

    @staticmethod
    def get_xpub(subpath='m'):
        # takes a string, like: m/44'/0'/23/23
        return b'xpub' + subpath.encode('ascii')

    @staticmethod
    def show_address(subpath, addr_fmt=AF_CLASSIC):
        # - takes a string, like: m/44'/0'/23/23
        # - shows on screen, no feedback from user expected
        assert not (addr_fmt & AFC_SCRIPT)
        return pack('<4sI', b'show', addr_fmt) + subpath.encode('ascii')

    @staticmethod
    def show_p2sh_address(M, xfp_paths, witdeem_script, addr_fmt=AF_P2SH):
        # For multisig (aka) P2SH cases, you will need all the info required to build
        # the redeem script, and the Coldcard must already have been enrolled 
        # into the wallet.
        # - redeem script must be provided
        # - full subkey paths for each involved key is required in a list of lists of ints, where
        #   is a XFP and derivation path, like in BIP174
        # - the order of xfp_paths must match the order of pubkeys in
        #   redeem script (after BIP67 sort). This allows for dup xfp values.
        assert addr_fmt & AFC_SCRIPT
        assert 30 <= len(witdeem_script) <= 520

        rv = pack('<4sIBBH', b'p2sh', addr_fmt, M, len(xfp_paths), len(witdeem_script))
        rv += witdeem_script

        for xfp_path in xfp_paths:
            ln = len(xfp_path)
            rv += pack('<B%dI' % ln, ln, *xfp_path)

        return rv

    @staticmethod
    def block_chain():
        # ask what blockchain it's set for; expect "BTC" or "XTN"
        return b'blkc'

    @staticmethod
    def sim_keypress(key):
        # Simulator ONLY: pretend a key is pressed
        return b'XKEY' + key

    @staticmethod
    def bag_number(new_number=b''):
        # one time only: put into bag, or readback bag
        return b'bagi' + bytes(new_number)

    @staticmethod
    def hsm_start(length=0, file_sha=b''):
        if length:
            # New policy already be uploaded as a JSON file, get approval and start.
            assert len(file_sha) == 32
            return pack('<4sI32s', b'hsms', length, file_sha)
        else:
            # Use policy on device already. Confirmation still required by local user.
            return b'hsms'

    @staticmethod
    def hsm_status():
        # get current status of HSM mode and/or policy defined already. Returns JSON
        return b'hsts'

    @staticmethod
    def create_user(username, auth_mode, secret=b''):
        # create username, with pre-shared secret/password, or we generate.
        # auth_model should be one of USER_AUTH_*
        # for TOTP/HOTP, secret can be empty. Set bit 0x80 in auth_mode and QR will be used
        assert 1 <= len(username) <= MAX_USERNAME_LEN
        assert len(secret) in { 0, 10, 20, 32}
        return pack('<4sBBB', b'nwur', auth_mode, len(username), len(secret)) + username + secret

    @staticmethod
    def delete_user(username):
        # remove a username and forget secret; cannot be used in HSM mode (only before)
        assert 0 < len(username) <= MAX_USERNAME_LEN
        return pack('<4sB', b'rmur', len(username)) + username

    @staticmethod
    def user_auth(username, token, totp_time=0):
        # HSM mode: try an authentication method for a username
        assert 0 < len(username) <= 16
        assert 6 <= len(token) <= 32
        return pack('<4sIBB', b'user', totp_time, len(username), len(token)) + username + token

    @staticmethod
    def get_storage_locker():
        # returns up to 414 bytes of user-defined sensitive data
        return b'gslr'


class CCProtocolUnpacker:
    # Take a binary response, and turn it into a python object
    # - we support a number of signatures, and expand as needed
    # - some will be general-purpose, but others can be very specific to one command
    # - given full rx message to work from
    # - this is done after un-framing

    @classmethod
    def decode(cls, msg):
        assert len(msg) >= 4
        sign = str(msg[0:4], 'utf8', 'ignore')

        d = getattr(cls, sign, cls)
        if d is cls:
            raise CCFramingError('Unknown response signature: ' + repr(sign))

        return d(msg)
        

    # struct info for each response
    
    def okay(msg):
        # trivial response, w/ no content
        assert len(msg) == 4
        return None

    # low-level errors
    def fram(msg):
        raise CCFramingError("Framing Error", str(msg[4:], 'utf8'))
    def err_(msg):
        raise CCProtoError("Coldcard Error: " + str(msg[4:], 'utf8', 'ignore'), msg[4:])

    def refu(msg):
        # user didn't want to approve something
        raise CCUserRefused()

    def busy(msg):
        # user didn't want to approve something
        raise CCBusyError()

    def biny(msg):
        # binary string: length implied by msg framing
        return msg[4:]

    def int1(msg):
        return unpack_from('<I', msg, 4)[0]

    def int2(msg):
        return unpack_from('<2I', msg, 4)

    def int3(msg):
        return unpack_from('<3I', msg, 4)

    def mypb(msg):
        # response to "ncry" command: 
        # - the (uncompressed) pubkey of the Coldcard
        # - info about master key: xpub, fingerprint of that
        # - anti-MitM: remote xpub 
        # session key is SHA256(point on sec256pk1 in binary) via D-H
        dev_pubkey, fingerprint, xpub_len = unpack_from('64sII', msg, 4)
        xpub = msg[-xpub_len:] if xpub_len else b''
        return dev_pubkey, fingerprint, xpub

    def asci(msg):
        # hex/base58 string or other for-computers string, which isn't international
        return msg[4:].decode('ascii')

    def smrx(msg):
        # message signing result. application specific!
        # returns actual address used (text), and raw binary signature (65 bytes)
        aln = unpack_from('<I', msg, 4)[0]
        return msg[8:aln+8].decode('ascii'), msg[8+aln:]

    def strx(msg):
        # txn signing result, or other file operation. application specific!
        # returns length of resulting PSBT and it's sha256
        ln, sha = unpack_from('<I32s', msg, 4)
        return ln, sha

# EOF



# Automatically generated by pb2py
# fmt: off
if __debug__:
    try:
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass

PAYTOADDRESS: Literal[0] = 0
PAYTOSCRIPTHASH: Literal[1] = 1
PAYTOMULTISIG: Literal[2] = 2
PAYTOOPRETURN: Literal[3] = 3
PAYTOWITNESS: Literal[4] = 4
PAYTOP2SHWITNESS: Literal[5] = 5



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'mainwindow.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        if not MainWindow.objectName():
            MainWindow.setObjectName(u"MainWindow")
        MainWindow.resize(828, 430)
        self.centralwidget = QWidget(MainWindow)
        self.centralwidget.setObjectName(u"centralwidget")
        self.verticalLayoutWidget = QWidget(self.centralwidget)
        self.verticalLayoutWidget.setObjectName(u"verticalLayoutWidget")
        self.verticalLayoutWidget.setGeometry(QRect(19, 29, 794, 375))
        self.verticalLayout = QVBoxLayout(self.verticalLayoutWidget)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalSpacer = QSpacerItem(200, 20, QSizePolicy.Maximum, QSizePolicy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer)

        self.enumerate_combobox = QComboBox(self.verticalLayoutWidget)
        self.enumerate_combobox.setObjectName(u"enumerate_combobox")
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.enumerate_combobox.sizePolicy().hasHeightForWidth())
        self.enumerate_combobox.setSizePolicy(sizePolicy)
        self.enumerate_combobox.setMinimumSize(QSize(0, 0))

        self.horizontalLayout.addWidget(self.enumerate_combobox)

        self.enumerate_refresh_button = QPushButton(self.verticalLayoutWidget)
        self.enumerate_refresh_button.setObjectName(u"enumerate_refresh_button")
        sizePolicy1 = QSizePolicy(QSizePolicy.Maximum, QSizePolicy.Fixed)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.enumerate_refresh_button.sizePolicy().hasHeightForWidth())
        self.enumerate_refresh_button.setSizePolicy(sizePolicy1)
        self.enumerate_refresh_button.setMaximumSize(QSize(100, 30))

        self.horizontalLayout.addWidget(self.enumerate_refresh_button)


        self.verticalLayout.addLayout(self.horizontalLayout)

        self.line = QFrame(self.verticalLayoutWidget)
        self.line.setObjectName(u"line")
        self.line.setFrameShape(QFrame.HLine)
        self.line.setFrameShadow(QFrame.Sunken)

        self.verticalLayout.addWidget(self.line)

        self.gridLayout = QGridLayout()
        self.gridLayout.setObjectName(u"gridLayout")
        self.sendpin_button = QPushButton(self.verticalLayoutWidget)
        self.sendpin_button.setObjectName(u"sendpin_button")
        self.sendpin_button.setEnabled(False)

        self.gridLayout.addWidget(self.sendpin_button, 1, 1, 1, 1)

        self.setpass_button = QPushButton(self.verticalLayoutWidget)
        self.setpass_button.setObjectName(u"setpass_button")

        self.gridLayout.addWidget(self.setpass_button, 1, 0, 1, 1)

        self.signtx_button = QPushButton(self.verticalLayoutWidget)
        self.signtx_button.setObjectName(u"signtx_button")
        self.signtx_button.setEnabled(False)

        self.gridLayout.addWidget(self.signtx_button, 2, 1, 1, 1)

        self.getxpub_button = QPushButton(self.verticalLayoutWidget)
        self.getxpub_button.setObjectName(u"getxpub_button")
        self.getxpub_button.setEnabled(False)

        self.gridLayout.addWidget(self.getxpub_button, 2, 0, 1, 1)

        self.signmsg_button = QPushButton(self.verticalLayoutWidget)
        self.signmsg_button.setObjectName(u"signmsg_button")
        self.signmsg_button.setEnabled(False)

        self.gridLayout.addWidget(self.signmsg_button, 2, 2, 1, 1)

        self.actions_label = QLabel(self.verticalLayoutWidget)
        self.actions_label.setObjectName(u"actions_label")
        sizePolicy2 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Maximum)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.actions_label.sizePolicy().hasHeightForWidth())
        self.actions_label.setSizePolicy(sizePolicy2)
        self.actions_label.setMaximumSize(QSize(100, 20))

        self.gridLayout.addWidget(self.actions_label, 0, 0, 1, 1)

        self.getkeypool_opts_button = QPushButton(self.verticalLayoutWidget)
        self.getkeypool_opts_button.setObjectName(u"getkeypool_opts_button")
        self.getkeypool_opts_button.setEnabled(False)

        self.gridLayout.addWidget(self.getkeypool_opts_button, 1, 2, 1, 1)

        self.display_addr_button = QPushButton(self.verticalLayoutWidget)
        self.display_addr_button.setObjectName(u"display_addr_button")
        self.display_addr_button.setEnabled(False)

        self.gridLayout.addWidget(self.display_addr_button, 3, 0, 1, 1)

        self.toggle_passphrase_button = QPushButton(self.verticalLayoutWidget)
        self.toggle_passphrase_button.setObjectName(u"toggle_passphrase_button")
        self.toggle_passphrase_button.setEnabled(False)

        self.gridLayout.addWidget(self.toggle_passphrase_button, 3, 1, 1, 1)


        self.verticalLayout.addLayout(self.gridLayout)

        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.keypool_label = QLabel(self.verticalLayoutWidget)
        self.keypool_label.setObjectName(u"keypool_label")

        self.horizontalLayout_2.addWidget(self.keypool_label)

        self.keypool_textedit = QPlainTextEdit(self.verticalLayoutWidget)
        self.keypool_textedit.setObjectName(u"keypool_textedit")
        self.keypool_textedit.setReadOnly(True)

        self.horizontalLayout_2.addWidget(self.keypool_textedit)


        self.verticalLayout.addLayout(self.horizontalLayout_2)

        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.desc_label = QLabel(self.verticalLayoutWidget)
        self.desc_label.setObjectName(u"desc_label")

        self.horizontalLayout_3.addWidget(self.desc_label)

        self.desc_textedit = QPlainTextEdit(self.verticalLayoutWidget)
        self.desc_textedit.setObjectName(u"desc_textedit")
        self.desc_textedit.setReadOnly(True)

        self.horizontalLayout_3.addWidget(self.desc_textedit)


        self.verticalLayout.addLayout(self.horizontalLayout_3)

        MainWindow.setCentralWidget(self.centralwidget)
        self.statusbar = QStatusBar(MainWindow)
        self.statusbar.setObjectName(u"statusbar")
        MainWindow.setStatusBar(self.statusbar)

        self.retranslateUi(MainWindow)

        QMetaObject.connectSlotsByName(MainWindow)
    # setupUi

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(QCoreApplication.translate("MainWindow", u"MainWindow", None))
        self.enumerate_refresh_button.setText(QCoreApplication.translate("MainWindow", u"Refresh", None))
        self.sendpin_button.setText(QCoreApplication.translate("MainWindow", u"Send Pin", None))
        self.setpass_button.setText(QCoreApplication.translate("MainWindow", u"Set Passphrase", None))
        self.signtx_button.setText(QCoreApplication.translate("MainWindow", u"Sign PSBT", None))
        self.getxpub_button.setText(QCoreApplication.translate("MainWindow", u"Get an xpub", None))
        self.signmsg_button.setText(QCoreApplication.translate("MainWindow", u"Sign Message", None))
        self.actions_label.setText(QCoreApplication.translate("MainWindow", u"Actions:", None))
#if QT_CONFIG(tooltip)
        self.getkeypool_opts_button.setToolTip(QCoreApplication.translate("MainWindow", u"Change the options used for getkeypool", None))
#endif // QT_CONFIG(tooltip)
        self.getkeypool_opts_button.setText(QCoreApplication.translate("MainWindow", u"Change getkeypool options", None))
        self.display_addr_button.setText(QCoreApplication.translate("MainWindow", u"Display Address", None))
        self.toggle_passphrase_button.setText(QCoreApplication.translate("MainWindow", u"Toggle Passphrase", None))
        self.keypool_label.setText(QCoreApplication.translate("MainWindow", u"Keypool:", None))
        self.desc_label.setText(QCoreApplication.translate("MainWindow", u"Descriptors:", None))
    # retranslateUi




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hwilib',
 'hwilib.devices',
 'hwilib.devices.btchip',
 'hwilib.devices.ckcc',
 'hwilib.devices.trezorlib',
 'hwilib.devices.trezorlib.messages',
 'hwilib.devices.trezorlib.transport',
 'hwilib.ui']

package_data = \
{'': ['*'], 'hwilib': ['udev/*']}

modules = \
['hwi', 'hwi-qt']
install_requires = \
['bitbox02>=5.2.0,<6.0.0',
 'ecdsa>=0,<1',
 'hidapi>=0,<1',
 'libusb1>=1.7,<2.0',
 'mnemonic>=0,<1',
 'pyaes>=1.6,<2.0',
 'typing-extensions>=3.7,<4.0']

extras_require = \
{'qt': ['pyside2>=5.14.0,<6.0.0']}

entry_points = \
{'console_scripts': ['hwi = hwilib._cli:main', 'hwi-qt = hwilib._gui:main']}

setup_kwargs = {
    'name': 'hwi',
    'version': '2.0.0',
    'description': 'A library for working with Bitcoin hardware wallets',
    'long_description': "# Bitcoin Hardware Wallet Interface\n\n[![Build Status](https://api.cirrus-ci.com/github/bitcoin-core/HWI.svg)](https://cirrus-ci.com/github/bitcoin-core/HWI)\n[![Documentation Status](https://readthedocs.org/projects/hwi/badge/?version=latest)](https://hwi.readthedocs.io/en/latest/?badge=latest)\n\nThe Bitcoin Hardware Wallet Interface is a Python library and command line tool for interacting with hardware wallets.\nIt provides a standard way for software to work with hardware wallets without needing to implement device specific drivers.\nPython software can use the provided library (`hwilib`). Software in other languages can execute the `hwi` tool.\n\nCaveat emptor: Inclusion of a specific hardware wallet vendor does not imply any endorsement of quality or security.\n\n## Prerequisites\n\nPython 3 is required. The libraries and [udev rules](hwilib/udev/README.md) for each device must also be installed. Some libraries will need to be installed\n\nFor Ubuntu/Debian:\n```\nsudo apt install libusb-1.0-0-dev libudev-dev python3-dev\n```\n\nFor Centos:\n```\nsudo yum -y install python3-devel libusbx-devel systemd-devel\n```\n\nFor macOS:\n```\nbrew install libusb\n```\n\n## Install\n\n```\ngit clone https://github.com/bitcoin-core/HWI.git\ncd HWI\npoetry install # or 'pip3 install .' or 'python3 setup.py install'\n```\n\nThis project uses the [Poetry](https://github.com/sdispater/poetry) dependency manager. HWI and its dependencies can be installed via poetry by executing the following in the root source directory:\n\n```\npoetry install\n```\n\nPip can also be used to automatically install HWI and its dependencies using the `setup.py` file (which is usually in sync with `pyproject.toml`):\n\n```\npip3 install .\n```\n\nThe `setup.py` file can be used to install HWI and its dependencies so long as `setuptools` is also installed:\n\n```\npip3 install -U setuptools\npython3 setup.py install\n```\n\n## Dependencies\n\nSee `pyproject.toml` for all dependencies. Dependencies under `[tool.poetry.dependecies]` are user dependencies, and `[tool.poetry.dev-dependencies]` for development based dependencies. These dependencies will be installed with any of the three above installation methods.\n\n## Usage\n\nTo use, first enumerate all devices and find the one that you want to use with\n\n```\n./hwi.py enumerate\n```\n\nOnce the device type and device path is known, issue commands to it like so:\n\n```\n./hwi.py -t <type> -d <path> <command> <command args>\n```\n\nAll output will be in JSON form and sent to `stdout`.\nAdditional information or prompts will be sent to `stderr` and will not necessarily be in JSON.\nThis additional information is for debugging purposes.\n\nTo see a complete list of available commands and global parameters, run\n`./hwi.py --help`.  To see options specific to a particular command,\npass the `--help` parameter after the command name; for example:\n\n```\n./hwi.py getdescriptors --help\n```\n\n## Documentation\n\nDocumentation for HWI can be found on [readthedocs.io](https://hwi.readthedocs.io/).\n\n### Device Support\n\nFor documentation on devices supported and how they are supported, please check the [device support page](https://hwi.readthedocs.io/en/latest/devices/index.html#support-matrix)\n\n### Using with Bitcoin Core\n\nSee [Using Bitcoin Core with Hardware Wallets](https://hwi.readthedocs.io/en/latest/examples/bitcoin-core-usage.html).\n\n## License\n\nThis project is available under the MIT License, Copyright Andrew Chow.\n",
    'author': 'Andrew Chow',
    'author_email': 'andrew@achow101.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/bitcoin-core/HWI',
    'packages': packages,
    'package_data': package_data,
    'py_modules': modules,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<3.10',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'sendpindialog.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_SendPinDialog(object):
    def setupUi(self, SendPinDialog):
        if not SendPinDialog.objectName():
            SendPinDialog.setObjectName(u"SendPinDialog")
        SendPinDialog.resize(257, 234)
        self.buttonBox = QDialogButtonBox(SendPinDialog)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setGeometry(QRect(60, 190, 181, 32))
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok)
        self.pin_lineedit = QLineEdit(SendPinDialog)
        self.pin_lineedit.setObjectName(u"pin_lineedit")
        self.pin_lineedit.setGeometry(QRect(10, 10, 231, 32))
        self.pin_lineedit.setReadOnly(False)
        self.gridLayoutWidget = QWidget(SendPinDialog)
        self.gridLayoutWidget.setObjectName(u"gridLayoutWidget")
        self.gridLayoutWidget.setGeometry(QRect(10, 50, 231, 131))
        self.gridLayout = QGridLayout(self.gridLayoutWidget)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.p2_button = QPushButton(self.gridLayoutWidget)
        self.p2_button.setObjectName(u"p2_button")

        self.gridLayout.addWidget(self.p2_button, 2, 1, 1, 1)

        self.p4_button = QPushButton(self.gridLayoutWidget)
        self.p4_button.setObjectName(u"p4_button")

        self.gridLayout.addWidget(self.p4_button, 1, 0, 1, 1)

        self.p1_button = QPushButton(self.gridLayoutWidget)
        self.p1_button.setObjectName(u"p1_button")

        self.gridLayout.addWidget(self.p1_button, 2, 0, 1, 1)

        self.p8_button = QPushButton(self.gridLayoutWidget)
        self.p8_button.setObjectName(u"p8_button")

        self.gridLayout.addWidget(self.p8_button, 0, 1, 1, 1)

        self.p7_button = QPushButton(self.gridLayoutWidget)
        self.p7_button.setObjectName(u"p7_button")

        self.gridLayout.addWidget(self.p7_button, 0, 0, 1, 1)

        self.p5_button = QPushButton(self.gridLayoutWidget)
        self.p5_button.setObjectName(u"p5_button")

        self.gridLayout.addWidget(self.p5_button, 1, 1, 1, 1)

        self.p9_button = QPushButton(self.gridLayoutWidget)
        self.p9_button.setObjectName(u"p9_button")

        self.gridLayout.addWidget(self.p9_button, 0, 2, 1, 1)

        self.p6_button = QPushButton(self.gridLayoutWidget)
        self.p6_button.setObjectName(u"p6_button")

        self.gridLayout.addWidget(self.p6_button, 1, 2, 1, 1)

        self.p3_button = QPushButton(self.gridLayoutWidget)
        self.p3_button.setObjectName(u"p3_button")

        self.gridLayout.addWidget(self.p3_button, 2, 2, 1, 1)


        self.retranslateUi(SendPinDialog)
        self.buttonBox.accepted.connect(SendPinDialog.accept)
        self.buttonBox.rejected.connect(SendPinDialog.reject)

        QMetaObject.connectSlotsByName(SendPinDialog)
    # setupUi

    def retranslateUi(self, SendPinDialog):
        SendPinDialog.setWindowTitle(QCoreApplication.translate("SendPinDialog", u"Dialog", None))
        self.p2_button.setText(QCoreApplication.translate("SendPinDialog", u"?", None))
        self.p4_button.setText(QCoreApplication.translate("SendPinDialog", u"?", None))
        self.p1_button.setText(QCoreApplication.translate("SendPinDialog", u"?", None))
        self.p8_button.setText(QCoreApplication.translate("SendPinDialog", u"?", None))
        self.p7_button.setText(QCoreApplication.translate("SendPinDialog", u"?", None))
        self.p5_button.setText(QCoreApplication.translate("SendPinDialog", u"?", None))
        self.p9_button.setText(QCoreApplication.translate("SendPinDialog", u"?", None))
        self.p6_button.setText(QCoreApplication.translate("SendPinDialog", u"?", None))
        self.p3_button.setText(QCoreApplication.translate("SendPinDialog", u"?", None))
    # retranslateUi




# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
        EnumTypeInputScriptType = Literal[0, 1, 2, 3, 4]
        EnumTypeAmountUnit = Literal[0, 1, 2, 3]
    except ImportError:
        pass


class AuthorizeCoinJoin(p.MessageType):
    MESSAGE_WIRE_TYPE = 51
    UNSTABLE = True

    def __init__(
        self,
        *,
        coordinator: str,
        max_total_fee: int,
        address_n: List[int] = None,
        fee_per_anonymity: int = None,
        coin_name: str = "Bitcoin",
        script_type: EnumTypeInputScriptType = 0,
        amount_unit: EnumTypeAmountUnit = 0,
    ) -> None:
        self.address_n = address_n if address_n is not None else []
        self.coordinator = coordinator
        self.max_total_fee = max_total_fee
        self.fee_per_anonymity = fee_per_anonymity
        self.coin_name = coin_name
        self.script_type = script_type
        self.amount_unit = amount_unit

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('coordinator', p.UnicodeType, p.FLAG_REQUIRED),
            2: ('max_total_fee', p.UVarintType, p.FLAG_REQUIRED),
            3: ('fee_per_anonymity', p.UVarintType, None),
            4: ('address_n', p.UVarintType, p.FLAG_REPEATED),
            5: ('coin_name', p.UnicodeType, "Bitcoin"),  # default=Bitcoin
            6: ('script_type', p.EnumType("InputScriptType", (0, 1, 2, 3, 4)), 0),  # default=SPENDADDRESS
            11: ('amount_unit', p.EnumType("AmountUnit", (0, 1, 2, 3)), 0),  # default=BITCOIN
        }



#! /usr/bin/env python3

"""
Commands
********

The functions in this module are the primary way to interact with hardware wallets.
Each function that takes a ``client`` uses a :class:`~hwilib.hwwclient.HardwareWalletClient`.
The functions then call public members of that client to retrieve the data needed.

Clients can be constructed using :func:`~find_device` or :func:`~get_client`.

The :func:`~enumerate` function returns information about what devices are available to be connected to.
These information can then be used with :func:`~find_device` or :func:`~get_client` to get a :class:`~hwilib.hwwclient.HardwareWalletClient`.

Note that this documentation does not specify every exception that can be raised.
Many exceptions are buried within the functions implemented by each device's :class:`~hwilib.hwwclient.HardwareWalletClient`.
For more information about the exceptions that those can raise, please see the specific client documentation.
"""

import importlib
import logging
import platform

from ._base58 import xpub_to_pub_hex
from .key import (
    get_bip44_purpose,
    get_bip44_chain,
    H_,
    HARDENED_FLAG,
    is_hardened,
    KeyOriginInfo,
    parse_path,
)
from .errors import (
    BadArgumentError,
    NotImplementedError,
    UnknownDeviceError,
    UnavailableActionError,
)
from .descriptor import (
    Descriptor,
    parse_descriptor,
    MultisigDescriptor,
    PKHDescriptor,
    PubkeyProvider,
    SHDescriptor,
    WPKHDescriptor,
    WSHDescriptor,
)
from .devices import __all__ as all_devs
from .common import (
    AddressType,
)
from .hwwclient import HardwareWalletClient
from .psbt import PSBT

from itertools import count
from typing import (
    Any,
    Dict,
    List,
    Optional,
)


py_enumerate = enumerate


# Get the client for the device
def get_client(device_type: str, device_path: str, password: str = "", expert: bool = False) -> Optional[HardwareWalletClient]:
    """
    Returns a HardwareWalletClient for the given device type at the device path

    :param device_type: The type of device
    :param device_path: The path specifying where the device can be accessed as returned by :func:`~enumerate`
    :param password: The password to use for this device
    :param expert: Whether the device should be opened in expert mode (prints more information for some commands)
    :return: A :class:`~hwilib.hwwclient.HardwareWalletClient` to interact with the device
    :raises: UnknownDeviceError: if the device type is not known by HWI
    """

    device_type = device_type.split('_')[0]
    class_name = device_type.capitalize()
    module = device_type.lower()

    client: Optional[HardwareWalletClient] = None
    try:
        imported_dev = importlib.import_module('.devices.' + module, __package__)
        client_constructor = getattr(imported_dev, class_name + 'Client')
        client = client_constructor(device_path, password, expert)
    except ImportError:
        if client:
            client.close()
        raise UnknownDeviceError('Unknown device type specified')

    return client

# Get a list of all available hardware wallets
def enumerate(password: str = "") -> List[Dict[str, Any]]:
    """
    Enumerate all of the devices that HWI can potentially access.

    :param password: The password to use for devices which take passwords from the host.
    :return: A list of devices for which clients can be created for.
    """

    result: List[Dict[str, Any]] = []

    for module in all_devs:
        try:
            imported_dev = importlib.import_module('.devices.' + module, __package__)
            result.extend(imported_dev.enumerate(password)) # type: ignore
        except ImportError as e:
            # Warn for ImportErrors, but largely ignore them to allow users not install
            # all device dependencies if only one or some devices are wanted.
            logging.warn(f"{e}, required for {module}. Ignore if you do not want this device.")
            pass
    return result

# Fingerprint or device type required
def find_device(
    password: str = "",
    device_type: Optional[str] = None,
    fingerprint: Optional[str] = None,
    expert: bool = False,
) -> Optional[HardwareWalletClient]:
    """
    Find a device from the device type or fingerprint and get a client to access it.
    This is used as an alternative to :func:`~get_client` if the device path is not known.

    :param password: A password that may be needed to access the device if it can take passwords from the host
    :param device_type: The type of device. The client returned will be for this type of device.
        If not provided, the fingerprint must be provided
    :param fingerprint: The fingerprint of the master public key for the device.
        The client returned will have a master public key fingerprint matching this.
        If not provided, device_type must be provided.
    :param expert: Whether the device should be opened in expert mode (enables additional output for some actions)
    :return: A client to interact with the found device
    """

    devices = enumerate(password)
    for d in devices:
        if device_type is not None and d['type'] != device_type and d['model'] != device_type:
            continue
        client = None
        try:
            assert isinstance(d["type"], str)
            assert isinstance(d["path"], str)
            client = get_client(d['type'], d['path'], password, expert)
            if client is None:
                raise Exception()

            if fingerprint:
                master_fpr = d.get('fingerprint', None)
                if master_fpr is None:
                    master_fpr = client.get_master_fingerprint().hex()

                if master_fpr != fingerprint:
                    client.close()
                    continue
            return client
        except Exception:
            if client:
                client.close()
            pass # Ignore things we wouldn't get fingerprints for
    return None

def getmasterxpub(client: HardwareWalletClient, addrtype: AddressType = AddressType.WIT, account: int = 0) -> Dict[str, str]:
    """
    Get the master extended public key from a client

    :param client: The client to interact with
    :return: A dictionary containing the public key at the ``m/44'/0'/0'`` derivation path.
        Returned as ``{"xpub": <xpub string>}``.
    """
    return {"xpub": client.get_master_xpub(addrtype, account).to_string()}

def signtx(client: HardwareWalletClient, psbt: str) -> Dict[str, str]:
    """
    Sign a Partially Signed Bitcoin Transaction (PSBT) with the client.

    :param client: The client to interact with
    :param psbt: The PSBT to sign
    :return: A dictionary containing the processed PSBT serialized in Base64.
        Returned as ``{"psbt": <base64 psbt string>}``.
    """
    # Deserialize the transaction
    tx = PSBT()
    tx.deserialize(psbt)
    return {"psbt": client.sign_tx(tx).serialize()}

def getxpub(client: HardwareWalletClient, path: str, expert: bool = False) -> Dict[str, Any]:
    """
    Get the master public key at a path from a client

    :param client: The client to interact with
    :param path: The derivation path for the public key to retrieve
    :param expert: Whether to provide more information intended for experts.
    :return: A dictionary containing the public key at the ``bip32_path``.
        With expert mode, the information contained within the xpub are decoded and displayed.
        Returned as ``{"xpub": <xpub string>}``.
    """
    xpub = client.get_pubkey_at_path(path)
    result: Dict[str, Any] = {"xpub": xpub.to_string()}
    if expert:
        result.update(xpub.get_printable_dict())
    return result

def signmessage(client: HardwareWalletClient, message: str, path: str) -> Dict[str, str]:
    """
    Sign a message using the key at the derivation path with the client.

    The message will be signed using the Bitcoin signed message standard used by Bitcoin Core.
    The message can be either a string which is then encoded to bytes, or bytes.

    :param client: The client to interact with
    :param message: The message to sign
    :param path: The derivation path for the key to sign with
    :return: A dictionary containing the signature.
        Returned as ``{"signature": <base64 signature string>}``.
    """
    return {"signature": client.sign_message(message, path)}

def getkeypool_inner(
    client: HardwareWalletClient,
    path: str,
    start: int,
    end: int,
    internal: bool = False,
    keypool: bool = True,
    account: int = 0,
    addr_type: AddressType = AddressType.WIT
) -> List[Dict[str, Any]]:
    """
    :meta private:

    Construct a single dictionary that specifies a single descriptor and the extra fields needed for ``importmulti`` or ``importdescriptors`` to import it.

    :param path: The derivation path for the key in the descriptor
    :param start: The start index of the range, inclusive
    :param end: The end index of the range, inclusive
    :param internal: Whether to specify this import is change
    :param keypool: Whether to specify this import should be added to the keypool
    :param account: The BIP 44 account to use if ``path`` is not specified
    :param addr_type: The type of address the descriptor should create
    """
    master_fpr = client.get_master_fingerprint()

    desc = getdescriptor(client, master_fpr, path, internal, addr_type, account, start, end)

    if not isinstance(desc, Descriptor):
        return desc

    this_import: Dict[str, Any] = {}

    this_import['desc'] = desc.to_string()
    this_import['range'] = [start, end]
    this_import['timestamp'] = 'now'
    this_import['internal'] = internal
    this_import['keypool'] = keypool
    this_import['active'] = keypool
    this_import['watchonly'] = True
    return [this_import]

def getdescriptor(
    client: HardwareWalletClient,
    master_fpr: bytes,
    path: Optional[str] = None,
    internal: bool = False,
    addr_type: AddressType = AddressType.WIT,
    account: int = 0,
    start: Optional[int] = None,
    end: Optional[int] = None
) -> Descriptor:
    """
    Get a descriptor from the client.

    :param client: The client to interact with
    :param master_fpr: The hex string for the master fingerprint of the device to use in the descriptor
    :param path: The derivation path for the xpub from which additional keys will be derived.
    :param internal: Whether the dictionary should indicate that the descriptor should be for change addresses
    :param addr_type: The type of address the descriptor should create
    :param account: The BIP 44 account to use if ``path`` is not specified
    :param start: The start of the range to import, inclusive
    :param end: The end of the range to import, inclusive
    :return: The descriptor constructed given the above arguments and key fetched from the device
    :raises: BadArgumentError: if an argument is malformed or missing.
    """
    is_wpkh = addr_type is AddressType.WIT
    is_sh_wpkh = addr_type is AddressType.SH_WIT

    parsed_path = []
    if not path:
        # Purpose
        parsed_path.append(H_(get_bip44_purpose(addr_type)))

        # Coin type
        parsed_path.append(H_(get_bip44_chain(client.chain)))

        # Account
        parsed_path.append(H_(account))

        # Receive or change
        if internal:
            parsed_path.append(1)
        else:
            parsed_path.append(0)
    else:
        if path[0] != "m":
            raise BadArgumentError("Path must start with m/")
        if path[-1] != "*":
            raise BadArgumentError("Path must end with /*")
        parsed_path = parse_path(path[:-2])

    # Find the last hardened derivation:
    for i, p in zip(count(len(parsed_path) - 1, -1), reversed(parsed_path)):
        if is_hardened(p):
            break
    i += 1

    origin = KeyOriginInfo(master_fpr, parsed_path[:i])
    path_base = origin.get_derivation_path()

    path_suffix = ""
    for p in parsed_path[i:]:
        hardened = is_hardened(p)
        p &= ~HARDENED_FLAG
        path_suffix += "/{}{}".format(p, "h" if hardened else "")
    path_suffix += "/*"

    # Get the key at the base
    if client.xpub_cache.get(path_base) is None:
        client.xpub_cache[path_base] = client.get_pubkey_at_path(path_base).to_string()

    pubkey = PubkeyProvider(origin, client.xpub_cache.get(path_base, ""), path_suffix)
    if is_wpkh:
        return WPKHDescriptor(pubkey)
    elif is_sh_wpkh:
        return SHDescriptor(WPKHDescriptor(pubkey))
    else:
        return PKHDescriptor(pubkey)

def getkeypool(
    client: HardwareWalletClient,
    path: str,
    start: int,
    end: int,
    internal: bool = False,
    keypool: bool = True,
    account: int = 0,
    addr_type: AddressType = AddressType.WIT,
    addr_all: bool = False
) -> List[Dict[str, Any]]:
    """
    Get a dictionary which can be passed to Bitcoin Core's ``importmulti`` or ``importdescriptors`` RPCs to import a watchonly wallet based on the client.
    By default, a descriptor for legacy addresses is returned.

    :param client: The client to interact with
    :param path: The derivation path for the xpub from which additional keys will be derived.
    :param start: The start of the range to import, inclusive
    :param end: The end of the range to import, inclusive
    :param internal: Whether the dictionary should indicate that the descriptor should be for change addresses
    :param keypool: Whether the dictionary should indicate that the dsecriptor should be added to the Bitcoin Core keypool/addresspool
    :param account: The BIP 44 account to use if ``path`` is not specified
    :param sh_wpkh: Whether to return a descriptor specifying p2sh-segwit addresses
    :param wpkh: Whether to return a descriptor specifying native segwit addresses
    :param addr_all: Whether to return a multiple descriptors for every address type
    :return: The dictionary containing the descriptor and all of the arguments for ``importmulti`` or ``importdescriptors``
    :raises: BadArgumentError: if an argument is malformed or missing.
    """

    addr_types = [addr_type]
    if addr_all:
        addr_types = list(AddressType)

    # When no specific path or internal-ness is specified, create standard types
    chains: List[Dict[str, Any]] = []
    if path is None and not internal:
        for addr_type in addr_types:
            for internal_addr in [False, True]:
                chains = chains + getkeypool_inner(client, None, start, end, internal_addr, keypool, account, addr_type)
        return chains
    else:
        assert len(addr_types) == 1
        return getkeypool_inner(client, path, start, end, internal, keypool, account, addr_types[0])


def getdescriptors(
    client: HardwareWalletClient,
    account: int = 0
) -> Dict[str, List[str]]:
    """
    Get descriptors from the client.

    :param client: The client to interact with
    :param account: The BIP 44 account to use
    :return: Multiple descriptors from the device matching the BIP 44 standard paths and the given ``account``.
    :raises: BadArgumentError: if an argument is malformed or missing.
    """
    master_fpr = client.get_master_fingerprint()

    result = {}

    for internal in [False, True]:
        descriptors = []
        for addr_type in (AddressType.LEGACY, AddressType.SH_WIT, AddressType.WIT):
            try:
                desc = getdescriptor(client, master_fpr=master_fpr, internal=internal, addr_type=addr_type, account=account)
            except UnavailableActionError:
                # Device does not support this address type or network. Skip.
                continue
            if not isinstance(desc, Descriptor):
                return desc
            descriptors.append(desc.to_string())
        if internal:
            result["internal"] = descriptors
        else:
            result["receive"] = descriptors

    return result

def displayaddress(
    client: HardwareWalletClient,
    path: Optional[str] = None,
    desc: Optional[str] = None,
    addr_type: AddressType = AddressType.WIT
) -> Dict[str, str]:
    """
    Display an address on the device for client.
    The address can be specified by the path with additional parameters, or by a descriptor.

    :param client: The client to interact with
    :param path: The path of the address to display. Mutually exclusive with ``desc``
    :param desc: The descriptor to display the address for. Mutually exclusive with ``path``
    :param addr_type: The address type to return. Only works with ``path``
    :return: A dictionary containing the address displayed.
        Returned as ``{"address": <base58 or bech32 address string>}``.
    :raises: BadArgumentError: if an argument is malformed, missing, or conflicts.
    """
    if path is not None:
        return {"address": client.display_singlesig_address(path, addr_type)}
    elif desc is not None:
        descriptor = parse_descriptor(desc)
        addr_type = AddressType.LEGACY
        is_sh = isinstance(descriptor, SHDescriptor)
        is_wsh = isinstance(descriptor, WSHDescriptor)
        if is_sh or is_wsh:
            assert descriptor.subdescriptor
            descriptor = descriptor.subdescriptor
            if isinstance(descriptor, WSHDescriptor):
                is_wsh = True
                assert descriptor.subdescriptor
                descriptor = descriptor.subdescriptor
            if isinstance(descriptor, MultisigDescriptor):
                if is_sh and is_wsh:
                    addr_type = AddressType.SH_WIT
                elif not is_sh and is_wsh:
                    addr_type = AddressType.WIT
                return {"address": client.display_multisig_address(addr_type, descriptor)}
        is_wpkh = isinstance(descriptor, WPKHDescriptor)
        if isinstance(descriptor, PKHDescriptor) or is_wpkh:
            pubkey = descriptor.pubkeys[0]
            if pubkey.origin is None:
                raise BadArgumentError(f"Descriptor missing origin info: {desc}")
            if pubkey.origin.fingerprint != client.get_master_fingerprint():
                raise BadArgumentError(f"Descriptor fingerprint does not match device: {desc}")
            xpub = client.get_pubkey_at_path(pubkey.origin.get_derivation_path()).to_string()
            if pubkey.pubkey != xpub and pubkey.pubkey != xpub_to_pub_hex(xpub):
                raise BadArgumentError(f"Key in descriptor does not match device: {desc}")
            if is_sh and is_wpkh:
                addr_type = AddressType.SH_WIT
            elif not is_sh and is_wpkh:
                addr_type = AddressType.WIT
            return {"address": client.display_singlesig_address(pubkey.get_full_derivation_path(0), addr_type)}
    raise BadArgumentError("Missing both path and descriptor")

def setup_device(client: HardwareWalletClient, label: str = "", backup_passphrase: str = "") -> Dict[str, bool]:
    """
    Setup a device that has not yet been initialized.

    :param client: The client to interact with
    :param label: The label to apply to the newly setup device
    :param backup_passphrase: The passphrase to use for the backup, if backups are encrypted for that device
    :return: A dictionary with the ``success`` key.
    """
    return {"success": client.setup_device(label, backup_passphrase)}

def wipe_device(client: HardwareWalletClient) -> Dict[str, bool]:
    """
    Wipe a device

    :param client: The client to interact with
    :return: A dictionary with the ``success`` key.
    """
    return {"success": client.wipe_device()}

def restore_device(client: HardwareWalletClient, label: str = "", word_count: int = 24) -> Dict[str, bool]:
    """
    Restore a backup to a device that has not yet been initialized.

    :param client: The client to interact with
    :param label: The label to apply to the newly setup device
    :param word_count: The number of words in the recovery phrase
    :return: A dictionary with the ``success`` key.
    """
    return {"success": client.restore_device(label, word_count)}

def backup_device(client: HardwareWalletClient, label: str = "", backup_passphrase: str = "") -> Dict[str, bool]:
    """
    Create a backup of the device

    :param client: The client to interact with
    :param label: The label to apply to the newly setup device
    :param backup_passphrase: The passphrase to use for the backup, if backups are encrypted for that device
    :return: A dictionary with the ``success`` key.
    """
    return {"success": client.backup_device(label, backup_passphrase)}

def prompt_pin(client: HardwareWalletClient) -> Dict[str, bool]:
    """
    Trigger the device to show the setup for PIN entry.

    :param client: The client to interact with
    :return: A dictionary with the ``success`` key.
    """
    return {"success": client.prompt_pin()}

def send_pin(client: HardwareWalletClient, pin: str) -> Dict[str, bool]:
    """
    Send a PIN to the device after :func:`prompt_pin` has been called.

    :param client: The client to interact with
    :param pin: The PIN to send
    :return: A dictionary with the ``success`` key.
    """
    return {"success": client.send_pin(pin)}

def toggle_passphrase(client: HardwareWalletClient) -> Dict[str, bool]:
    """
    Toggle whether the device is using a BIP 39 passphrase.

    :param client: The client to interact with
    :return: A dictionary with the ``success`` key.
    """
    return {"success": client.toggle_passphrase()}

def install_udev_rules(source: str, location: str) -> Dict[str, bool]:
    """
    Install the udev rules to the local machine.
    The rules will be copied from the source to the location.
    ``udevadm`` will also be triggered and the rules reloaded so that the devices can be plugged in and used immediately.
    A ``plugdev`` group will also be created if it does not exist and the user will be added to it.

    The recommended source location is ``hwilib/udev``. The recommended destination location is ``/etc/udev/rules.d``

    This function is equivalent to::

        sudo cp hwilib/udev/*rules /etc/udev/rules.d/
        sudo udevadm trigger
        sudo udevadm control --reload-rules
        sudo groupadd plugdev
        sudo usermod -aG plugdev `whoami`

    :param source: The directory containing the udev rules to install
    :param location: The directory to install the udev rules to
    :return: A dictionary with the ``success`` key.
    :raises: NotImplementedError: if udev rules cannot be installed on this system, i.e. it is not linux.
    """
    if platform.system() == "Linux":
        from .udevinstaller import UDevInstaller
        return {"success": UDevInstaller.install(source, location)}
    raise NotImplementedError("udev rules are not needed on your platform")



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

if __debug__:
    try:
        from typing import Dict, List  # noqa: F401
        from typing_extensions import Literal  # noqa: F401
    except ImportError:
        pass


class PinMatrixAck(p.MessageType):
    MESSAGE_WIRE_TYPE = 19

    def __init__(
        self,
        *,
        pin: str,
    ) -> None:
        self.pin = pin

    @classmethod
    def get_fields(cls) -> Dict:
        return {
            1: ('pin', p.UnicodeType, p.FLAG_REQUIRED),
        }



#
# Constants and various "limits" shared between embedded and desktop USB protocol
#
try:
    from micropython import const
except ImportError:
    const = int

# For upload/download this is the max size of the data block.
MAX_BLK_LEN = const(2048)

# Max total message length, excluding framing overhead (1 byte per 64).
# - includes args for upload command
MAX_MSG_LEN = const(4+4+4+MAX_BLK_LEN)

# Max PSBT txn we support (384k bytes as PSBT)
# - the max on the wire for mainnet is 100k
# - but a PSBT might contain a full txn for each input
MAX_TXN_LEN = const(384*1024)

# Max size of any upload (firmware.dfu files in particular)
MAX_UPLOAD_LEN = const(2*MAX_TXN_LEN)

# Max length of text messages for signing
MSG_SIGNING_MAX_LENGTH = const(240)

# Types of user auth we support
USER_AUTH_TOTP = const(1)       # RFC6238
USER_AUTH_HOTP = const(2)       # RFC4226
USER_AUTH_HMAC = const(3)       # PBKDF2('hmac-sha256', secret, sha256(psbt), PBKDF2_ITER_COUNT)
USER_AUTH_SHOW_QR = const(0x80) # show secret on Coldcard screen (best for TOTP enroll)

MAX_USERNAME_LEN = 16
PBKDF2_ITER_COUNT = 2500

# Max depth for derived keys, in PSBT files, and USB commands
MAX_PATH_DEPTH = const(12)

# Bitmask used in sign_transaction (stxn) command
STXN_FINALIZE       = const(0x01)
STXN_VISUALIZE      = const(0x02)
STXN_SIGNED         = const(0x04)
STXN_FLAGS_MASK     = const(0x07)

# Bit values for address types
AFC_PUBKEY      = const(0x01)       # pay to hash of pubkey
AFC_SEGWIT      = const(0x02)       # requires a witness to spend
AFC_BECH32      = const(0x04)       # just how we're encoding it?
AFC_SCRIPT      = const(0x08)       # paying into a script
AFC_WRAPPED     = const(0x10)       # for transition/compat types for segwit vs. old

# Numeric codes for specific address types
AF_CLASSIC      = AFC_PUBKEY          # 1addr
AF_P2SH         = AFC_SCRIPT          # classic multisig / simple P2SH / 3hash
AF_P2WPKH       = AFC_PUBKEY  | AFC_SEGWIT | AFC_BECH32     # bc1qsdklfj
AF_P2WSH        = AFC_SCRIPT  | AFC_SEGWIT | AFC_BECH32     # segwit multisig
AF_P2WPKH_P2SH  = AFC_WRAPPED | AFC_PUBKEY | AFC_SEGWIT     # looks classic P2SH, but p2wpkh inside
AF_P2WSH_P2SH   = AFC_WRAPPED | AFC_SCRIPT | AFC_SEGWIT     # looks classic P2SH, segwit multisig

SUPPORTED_ADDR_FORMATS = frozenset([
    AF_CLASSIC,
    AF_P2SH,
    AF_P2WPKH,
    AF_P2WSH,
    AF_P2WPKH_P2SH,
    AF_P2WSH_P2SH,
])

# BIP-174 aka PSBT defined values
#
PSBT_GLOBAL_UNSIGNED_TX     = const(0)
PSBT_GLOBAL_XPUB            = const(1)

PSBT_IN_NON_WITNESS_UTXO    = const(0)
PSBT_IN_WITNESS_UTXO        = const(1)
PSBT_IN_PARTIAL_SIG         = const(2)
PSBT_IN_SIGHASH_TYPE        = const(3)
PSBT_IN_REDEEM_SCRIPT       = const(4)
PSBT_IN_WITNESS_SCRIPT      = const(5)
PSBT_IN_BIP32_DERIVATION    = const(6)
PSBT_IN_FINAL_SCRIPTSIG     = const(7)
PSBT_IN_FINAL_SCRIPTWITNESS = const(8)

PSBT_OUT_REDEEM_SCRIPT      = const(0)
PSBT_OUT_WITNESS_SCRIPT     = const(1)
PSBT_OUT_BIP32_DERIVATION   = const(2)

# EOF



"""
Coldcard
********
"""

from typing import (
    Dict,
    List,
    Union,
)

from ..descriptor import MultisigDescriptor
from ..hwwclient import HardwareWalletClient
from ..errors import (
    ActionCanceledError,
    BadArgumentError,
    DeviceBusyError,
    DeviceFailureError,
    UnavailableActionError,
    common_err_msgs,
    handle_errors,
)
from .ckcc.client import (
    ColdcardDevice,
    COINKITE_VID,
    CKCC_PID,
)
from .ckcc.protocol import (
    CCProtocolPacker,
    CCBusyError,
    CCProtoError,
    CCUserRefused,
)
from .ckcc.constants import (
    MAX_BLK_LEN,
    AF_P2WPKH,
    AF_CLASSIC,
    AF_P2WPKH_P2SH,
    AF_P2WSH,
    AF_P2SH,
    AF_P2WSH_P2SH,
)
from .._base58 import (
    get_xpub_fingerprint,
)
from ..key import (
    ExtendedKey,
)
from ..psbt import (
    PSBT,
)
from ..common import (
    AddressType,
    Chain,
)
from functools import wraps
from hashlib import sha256

import base64
import hid
import io
import sys
import time
import struct

from binascii import b2a_hex
from typing import (
    Any,
    Callable,
)

CC_SIMULATOR_SOCK = '/tmp/ckcc-simulator.sock'
# Using the simulator: https://github.com/Coldcard/firmware/blob/master/unix/README.md


def coldcard_exception(f: Callable[..., Any]) -> Callable[..., Any]:
    @wraps(f)
    def func(*args: Any, **kwargs: Any) -> Any:
        try:
            return f(*args, **kwargs)
        except CCProtoError as e:
            raise BadArgumentError(str(e))
        except CCUserRefused:
            raise ActionCanceledError('{} canceled'.format(f.__name__))
        except CCBusyError as e:
            raise DeviceBusyError(str(e))
    return func

# This class extends the HardwareWalletClient for ColdCard specific things
class ColdcardClient(HardwareWalletClient):

    def __init__(self, path: str, password: str = "", expert: bool = False) -> None:
        super(ColdcardClient, self).__init__(path, password, expert)
        # Simulator hard coded pipe socket
        if path == CC_SIMULATOR_SOCK:
            self.device = ColdcardDevice(sn=path)
        else:
            device = hid.device()
            device.open_path(path.encode())
            self.device = ColdcardDevice(dev=device)

    @coldcard_exception
    def get_pubkey_at_path(self, path: str) -> ExtendedKey:
        self.device.check_mitm()
        path = path.replace('h', '\'')
        path = path.replace('H', '\'')
        xpub_str = self.device.send_recv(CCProtocolPacker.get_xpub(path), timeout=None)
        xpub = ExtendedKey.deserialize(xpub_str)
        if self.chain != Chain.MAIN:
            xpub.version = ExtendedKey.TESTNET_PUBLIC
        return xpub

    def get_master_fingerprint(self) -> bytes:
        # quick method to get fingerprint of wallet
        return struct.pack('<I', self.device.master_fingerprint)

    @coldcard_exception
    def sign_tx(self, tx: PSBT) -> PSBT:
        """
        Sign a transaction with the Coldcard.

        - The Coldcard firmware only supports signing single key and multisig transactions. It cannot sign arbitrary scripts.
        - Multisigs need to be registered on the device before a transaction spending that multisig will be signed by the device.
        - Multisigs must use BIP 67. This can be accomplished in Bitcoin Core using the `sortedmulti()` descriptor, available in Bitcoin Core 0.20.
        """
        self.device.check_mitm()

        # Get this devices master key fingerprint
        xpub = self.device.send_recv(CCProtocolPacker.get_xpub('m/0\''), timeout=None)
        master_fp = get_xpub_fingerprint(xpub)

        # For multisigs, we may need to do multiple passes if we appear in an input multiple times
        passes = 1
        for psbt_in in tx.inputs:
            our_keys = 0
            for key in psbt_in.hd_keypaths.keys():
                keypath = psbt_in.hd_keypaths[key]
                if keypath.fingerprint == master_fp and key not in psbt_in.partial_sigs:
                    our_keys += 1
            if our_keys > passes:
                passes = our_keys

        for _ in range(passes):
            # Get psbt in hex and then make binary
            fd = io.BytesIO(base64.b64decode(tx.serialize()))

            # learn size (portable way)
            sz = fd.seek(0, 2)
            fd.seek(0)

            left = sz
            chk = sha256()
            for pos in range(0, sz, MAX_BLK_LEN):
                here = fd.read(min(MAX_BLK_LEN, left))
                if not here:
                    break
                left -= len(here)
                result = self.device.send_recv(CCProtocolPacker.upload(pos, sz, here))
                assert result == pos
                chk.update(here)

            # do a verify
            expect = chk.digest()
            result = self.device.send_recv(CCProtocolPacker.sha256())
            assert len(result) == 32
            if result != expect:
                raise DeviceFailureError("Wrong checksum:\nexpect: %s\n   got: %s" % (b2a_hex(expect).decode('ascii'), b2a_hex(result).decode('ascii')))

            # start the signing process
            ok = self.device.send_recv(CCProtocolPacker.sign_transaction(sz, expect), timeout=None)
            assert ok is None
            if self.device.is_simulator:
                self.device.send_recv(CCProtocolPacker.sim_keypress(b'y'))

            print("Waiting for OK on the Coldcard...", file=sys.stderr)

            while 1:
                time.sleep(0.250)
                done = self.device.send_recv(CCProtocolPacker.get_signed_txn(), timeout=None)
                if done is None:
                    continue
                break

            if len(done) != 2:
                raise DeviceFailureError('Failed: %r' % done)

            result_len, result_sha = done

            result = self.device.download_file(result_len, result_sha, file_number=1)

            tx = PSBT()
            tx.deserialize(base64.b64encode(result).decode())

        return tx

    @coldcard_exception
    def sign_message(self, message: Union[str, bytes], keypath: str) -> str:
        self.device.check_mitm()
        keypath = keypath.replace('h', '\'')
        keypath = keypath.replace('H', '\'')

        msg = message
        if not isinstance(message, bytes):
            msg = message.encode()
        ok = self.device.send_recv(
            CCProtocolPacker.sign_message(msg, keypath, AF_CLASSIC), timeout=None
        )
        assert ok is None
        if self.device.is_simulator:
            self.device.send_recv(CCProtocolPacker.sim_keypress(b'y'))

        while 1:
            time.sleep(0.250)
            done = self.device.send_recv(CCProtocolPacker.get_signed_msg(), timeout=None)
            if done is None:
                continue

            break

        if len(done) != 2:
            raise DeviceFailureError('Failed: %r' % done)

        _, raw = done

        sig = str(base64.b64encode(raw), 'ascii').replace('\n', '')
        return sig

    @coldcard_exception
    def display_singlesig_address(
        self,
        keypath: str,
        addr_type: AddressType,
    ) -> str:
        self.device.check_mitm()
        keypath = keypath.replace('h', '\'')
        keypath = keypath.replace('H', '\'')

        if addr_type == AddressType.SH_WIT:
            addr_fmt = AF_P2WPKH_P2SH
        elif addr_type == AddressType.WIT:
            addr_fmt = AF_P2WPKH
        elif addr_type == AddressType.LEGACY:
            addr_fmt = AF_CLASSIC
        else:
            raise BadArgumentError("Unknown address type")

        payload = CCProtocolPacker.show_address(keypath, addr_fmt=addr_fmt)

        address = self.device.send_recv(payload, timeout=None)
        assert isinstance(address, str)

        if self.device.is_simulator:
            self.device.send_recv(CCProtocolPacker.sim_keypress(b'y'))
        return address

    @coldcard_exception
    def display_multisig_address(
        self,
        addr_type: AddressType,
        multisig: MultisigDescriptor,
    ) -> str:
        if not multisig.is_sorted:
            raise BadArgumentError("Coldcards only allow sortedmulti descriptors")

        self.device.check_mitm()

        if addr_type == AddressType.SH_WIT:
            addr_fmt = AF_P2WSH_P2SH
        elif addr_type == AddressType.WIT:
            addr_fmt = AF_P2WSH
        elif addr_type == AddressType.LEGACY:
            addr_fmt = AF_P2SH
        else:
            raise BadArgumentError("Unknown address type")

        if not 1 <= len(multisig.pubkeys) <= 15:
            raise BadArgumentError("Must provide 1 to 15 keypaths to display a multisig address")

        redeem_script = (80 + int(multisig.thresh)).to_bytes(1, byteorder="little")

        if not 1 <= multisig.thresh <= len(multisig.pubkeys):
            raise BadArgumentError("Either the redeem script provided is invalid or the keypaths provided are insufficient")

        xfp_paths = []
        sorted_keys = sorted(zip([p.get_pubkey_bytes(0) for p in multisig.pubkeys], multisig.pubkeys))
        for pk, p in sorted_keys:
            xfp_paths.append(p.get_full_derivation_int_list(0))
            redeem_script += len(pk).to_bytes(1, byteorder="little") + pk

        redeem_script += (80 + len(multisig.pubkeys)).to_bytes(1, byteorder="little")
        redeem_script += b"\xae"

        payload = CCProtocolPacker.show_p2sh_address(multisig.thresh, xfp_paths, redeem_script, addr_fmt=addr_fmt)

        address = self.device.send_recv(payload, timeout=None)
        assert isinstance(address, str)

        if self.device.is_simulator:
            self.device.send_recv(CCProtocolPacker.sim_keypress(b'y'))
        return address

    def setup_device(self, label: str = "", passphrase: str = "") -> bool:
        """
        The Coldcard does not support setup via software.

        :raises UnavailableActionError: Always, this function is unavailable
        """
        raise UnavailableActionError('The Coldcard does not support software setup')

    def wipe_device(self) -> bool:
        """
        The Coldcard does not support wiping via software.

        :raises UnavailableActionError: Always, this function is unavailable
        """
        raise UnavailableActionError('The Coldcard does not support wiping via software')

    def restore_device(self, label: str = "", word_count: int = 24) -> bool:
        """
        The Coldcard does not support restoring via software.

        :raises UnavailableActionError: Always, this function is unavailable
        """
        raise UnavailableActionError('The Coldcard does not support restoring via software')

    @coldcard_exception
    def backup_device(self, label: str = "", passphrase: str = "") -> bool:
        """
        Creates a backup file in the current working directory. This file is protected by the
        passphrase shown on the Coldcard.

        :param label: Value is ignored
        :param passphrase: Value is ignored
        """
        self.device.check_mitm()

        ok = self.device.send_recv(CCProtocolPacker.start_backup())
        assert ok is None
        if self.device.is_simulator:
            self.device.send_recv(CCProtocolPacker.sim_keypress(b'y'))

        while 1:
            if self.device.is_simulator: # For the simulator, work through the password quiz. Eventually pressing 1 will work
                self.device.send_recv(CCProtocolPacker.sim_keypress(b'1'))

            time.sleep(0.250)
            done = self.device.send_recv(CCProtocolPacker.get_backup_file(), timeout=None)
            if done is None:
                continue
            break

        if len(done) != 2:
            raise DeviceFailureError('Failed: %r' % done)

        result_len, result_sha = done

        result = self.device.download_file(result_len, result_sha, file_number=0)
        filename = time.strftime('backup-%Y%m%d-%H%M.7z')
        open(filename, 'wb').write(result)
        return True

    def close(self) -> None:
        self.device.close()

    def prompt_pin(self) -> bool:
        """
        The Coldcard does not need a PIN sent from the host.

        :raises UnavailableActionError: Always, this function is unavailable
        """
        raise UnavailableActionError('The Coldcard does not need a PIN sent from the host')

    def send_pin(self, pin: str) -> bool:
        """
        The Coldcard does not need a PIN sent from the host.

        :raises UnavailableActionError: Always, this function is unavailable
        """
        raise UnavailableActionError('The Coldcard does not need a PIN sent from the host')

    def toggle_passphrase(self) -> bool:
        """
        The Coldcard does not support toggling passphrase from the host.

        :raises UnavailableActionError: Always, this function is unavailable
        """
        raise UnavailableActionError('The Coldcard does not support toggling passphrase from the host')

def enumerate(password: str = "") -> List[Dict[str, Any]]:
    results = []
    devices = hid.enumerate(COINKITE_VID, CKCC_PID)
    devices.append({'path': CC_SIMULATOR_SOCK.encode()})
    for d in devices:
        d_data: Dict[str, Any] = {}

        path = d['path'].decode()
        d_data['type'] = 'coldcard'
        d_data['model'] = 'coldcard'
        d_data['path'] = path
        d_data['needs_pin_sent'] = False
        d_data['needs_passphrase_sent'] = False

        if path == CC_SIMULATOR_SOCK:
            d_data['model'] += '_simulator'

        client = None
        with handle_errors(common_err_msgs["enumerate"], d_data):
            try:
                client = ColdcardClient(path)
                d_data['fingerprint'] = client.get_master_fingerprint().hex()
            except RuntimeError as e:
                # Skip the simulator if it's not there
                if str(e) == 'Cannot connect to simulator. Is it running?':
                    continue
                else:
                    raise e

        if client:
            client.close()

        results.append(d_data)

    return results



#! /usr/bin/env python3

import json
import logging
import sys
import time
from typing import Callable

from . import commands, __version__
from ._cli import HWIArgumentParser
from .errors import handle_errors, DEVICE_NOT_INITIALIZED
from .common import AddressType, Chain

try:
    from .ui.ui_bitbox02pairing import Ui_BitBox02PairingDialog
    from .ui.ui_displayaddressdialog import Ui_DisplayAddressDialog
    from .ui.ui_getxpubdialog import Ui_GetXpubDialog
    from .ui.ui_getkeypooloptionsdialog import Ui_GetKeypoolOptionsDialog
    from .ui.ui_mainwindow import Ui_MainWindow
    from .ui.ui_sendpindialog import Ui_SendPinDialog
    from .ui.ui_setpassphrasedialog import Ui_SetPassphraseDialog
    from .ui.ui_signmessagedialog import Ui_SignMessageDialog
    from .ui.ui_signpsbtdialog import Ui_SignPSBTDialog
except ImportError:
    print('Could not import UI files, did you run contrib/generate-ui.sh')
    exit(-1)

from PySide2.QtGui import QRegExpValidator
from PySide2.QtWidgets import QApplication, QDialog, QDialogButtonBox, QLineEdit, QMessageBox, QMainWindow
from PySide2.QtCore import QCoreApplication, QRegExp, Signal, Slot

import bitbox02.util

def do_command(f, *args, **kwargs):
    result = {}
    with handle_errors(result=result):
        result = f(*args, **kwargs)
    if 'error' in result:
        msg = 'Error: {}\nCode:{}'.format(result['error'], result['code'])
        QMessageBox.critical(None, "An Error Occurred", msg)
        return None
    return result

class SetPassphraseDialog(QDialog):
    def __init__(self):
        super(SetPassphraseDialog, self).__init__()
        self.ui = Ui_SetPassphraseDialog()
        self.ui.setupUi(self)
        self.setWindowTitle('Set Passphrase')

        self.ui.passphrase_lineedit.setFocus()

class SendPinDialog(QDialog):
    pin_sent_success = Signal()

    def __init__(self, client, prompt_pin=True):
        super(SendPinDialog, self).__init__()
        self.ui = Ui_SendPinDialog()
        self.ui.setupUi(self)
        self.setWindowTitle('Send Pin')
        self.client = client
        self.ui.pin_lineedit.setFocus()
        self.ui.pin_lineedit.setValidator(QRegExpValidator(QRegExp("[1-9]+"), None))
        self.ui.pin_lineedit.setEchoMode(QLineEdit.Password)

        self.ui.p1_button.clicked.connect(self.button_clicked(1))
        self.ui.p2_button.clicked.connect(self.button_clicked(2))
        self.ui.p3_button.clicked.connect(self.button_clicked(3))
        self.ui.p4_button.clicked.connect(self.button_clicked(4))
        self.ui.p5_button.clicked.connect(self.button_clicked(5))
        self.ui.p6_button.clicked.connect(self.button_clicked(6))
        self.ui.p7_button.clicked.connect(self.button_clicked(7))
        self.ui.p8_button.clicked.connect(self.button_clicked(8))
        self.ui.p9_button.clicked.connect(self.button_clicked(9))

        self.accepted.connect(self.sendpindialog_accepted)
        if prompt_pin:
            do_command(commands.prompt_pin, self.client)

    def button_clicked(self, number):
        @Slot()
        def button_clicked_num():
            self.ui.pin_lineedit.setText(self.ui.pin_lineedit.text() + str(number))
        return button_clicked_num

    @Slot()
    def sendpindialog_accepted(self):
        pin = self.ui.pin_lineedit.text()

        # Send the pin
        do_command(commands.send_pin, self.client, pin)
        self.client.close()
        self.client = None
        self.pin_sent_success.emit()

class GetXpubDialog(QDialog):
    def __init__(self, client):
        super(GetXpubDialog, self).__init__()
        self.ui = Ui_GetXpubDialog()
        self.ui.setupUi(self)
        self.setWindowTitle('Get xpub')
        self.client = client

        self.ui.path_lineedit.setValidator(QRegExpValidator(QRegExp("m(/[0-9]+['Hh]?)+"), None))
        self.ui.path_lineedit.setFocus()
        self.ui.buttonBox.button(QDialogButtonBox.Close).setAutoDefault(False)

        self.ui.getxpub_button.clicked.connect(self.getxpub_button_clicked)
        self.ui.buttonBox.clicked.connect(self.accept)

    @Slot()
    def getxpub_button_clicked(self):
        path = self.ui.path_lineedit.text()
        res = do_command(commands.getxpub, self.client, path)
        self.ui.xpub_lineedit.setText(res['xpub'])

class SignPSBTDialog(QDialog):
    def __init__(self, client):
        super(SignPSBTDialog, self).__init__()
        self.ui = Ui_SignPSBTDialog()
        self.ui.setupUi(self)
        self.setWindowTitle('Sign PSBT')
        self.client = client

        self.ui.psbt_in_textedit.setFocus()

        self.ui.sign_psbt_button.clicked.connect(self.sign_psbt_button_clicked)
        self.ui.buttonBox.clicked.connect(self.accept)

    @Slot()
    def sign_psbt_button_clicked(self):
        psbt_str = self.ui.psbt_in_textedit.toPlainText()
        res = do_command(commands.signtx, self.client, psbt_str)
        self.ui.psbt_out_textedit.setPlainText(res['psbt'])

class SignMessageDialog(QDialog):
    def __init__(self, client):
        super(SignMessageDialog, self).__init__()
        self.ui = Ui_SignMessageDialog()
        self.ui.setupUi(self)
        self.setWindowTitle('Sign Message')
        self.client = client

        self.ui.path_lineedit.setValidator(QRegExpValidator(QRegExp("m(/[0-9]+['Hh]?)+"), None))
        self.ui.msg_textedit.setFocus()

        self.ui.signmsg_button.clicked.connect(self.signmsg_button_clicked)
        self.ui.buttonBox.clicked.connect(self.accept)

    @Slot()
    def signmsg_button_clicked(self):
        msg_str = self.ui.msg_textedit.toPlainText()
        path = self.ui.path_lineedit.text()
        res = do_command(commands.signmessage, self.client, msg_str, path)
        self.ui.sig_textedit.setPlainText(res['signature'])

class DisplayAddressDialog(QDialog):
    def __init__(self, client):
        super(DisplayAddressDialog, self).__init__()
        self.ui = Ui_DisplayAddressDialog()
        self.ui.setupUi(self)
        self.setWindowTitle('Display Address')
        self.client = client

        self.ui.path_lineedit.setValidator(QRegExpValidator(QRegExp("m(/[0-9]+['Hh]?)+"), None))
        self.ui.path_lineedit.setFocus()

        self.ui.go_button.clicked.connect(self.go_button_clicked)
        self.ui.buttonBox.clicked.connect(self.accept)

    @Slot()
    def go_button_clicked(self):
        path = self.ui.path_lineedit.text()
        if self.ui.sh_wpkh_radio.isChecked():
            addrtype = AddressType.SH_WIT
        elif self.ui.wpkh_radio.isChecked():
            addrtype = AddressType.WIT
        elif self.ui.pkh_radio.isChecked():
            addrtype = AddressType.LEGACY
        else:
            assert False # How did this happen?
        res = do_command(commands.displayaddress, self.client, path, addr_type=addrtype)
        self.ui.address_lineedit.setText(res['address'])

class GetKeypoolOptionsDialog(QDialog):
    def __init__(self, opts):
        super(GetKeypoolOptionsDialog, self).__init__()
        self.ui = Ui_GetKeypoolOptionsDialog()
        self.ui.setupUi(self)
        self.setWindowTitle('Set getkeypool options')

        self.ui.start_spinbox.setValue(opts['start'])
        self.ui.end_spinbox.setValue(opts['end'])
        self.ui.internal_checkbox.setChecked(opts['internal'])
        self.ui.keypool_checkbox.setChecked(opts['keypool'])
        self.ui.account_spinbox.setValue(opts['account'])
        self.ui.path_lineedit.setValidator(QRegExpValidator(QRegExp("m(/[0-9]+['Hh]?)+"), None))
        if opts['account_used']:
            self.ui.account_radio.setChecked(True)
            self.ui.path_radio.setChecked(False)
            self.ui.path_lineedit.setEnabled(False)
            self.ui.account_spinbox.setEnabled(True)
            self.ui.account_spinbox.setValue(opts['account'])
        else:
            self.ui.account_radio.setChecked(False)
            self.ui.path_radio.setChecked(True)
            self.ui.path_lineedit.setEnabled(True)
            self.ui.account_spinbox.setEnabled(False)
            self.ui.path_lineedit.setText(opts['path'])
        self.ui.sh_wpkh_radio.setChecked(opts['addrtype'] == AddressType.SH_WIT)
        self.ui.wpkh_radio.setChecked(opts['addrtype'] == AddressType.WIT)
        self.ui.pkh_radio.setChecked(opts['addrtype'] == AddressType.LEGACY)

        self.ui.account_radio.toggled.connect(self.toggle_account)

    @Slot()
    def toggle_account(self, checked):
        if checked:
            self.ui.path_lineedit.setEnabled(False)
            self.ui.account_spinbox.setEnabled(True)
        else:
            self.ui.path_lineedit.setEnabled(True)
            self.ui.account_spinbox.setEnabled(False)

class BitBox02PairingDialog(QDialog):
    def __init__(self, pairing_code: str, device_response: Callable[[], bool]):
        super(BitBox02PairingDialog, self).__init__()
        self.ui = Ui_BitBox02PairingDialog()
        self.ui.setupUi(self)
        self.setWindowTitle('Verify BitBox02 pairing code')
        self.ui.pairingCode.setText(pairing_code.replace("\n", "<br>"))
        self.ui.buttonBox.setEnabled(False)
        self.device_response = device_response
        self.painted = False

    def paintEvent(self, ev):
        super().paintEvent(ev)
        self.painted = True

    def enable_buttons(self):
        self.ui.buttonBox.setEnabled(True)

class BitBox02NoiseConfig(bitbox02.util.BitBoxAppNoiseConfig):
    """ GUI elements to perform the BitBox02 pairing and attestatoin check """

    def show_pairing(self, code: str, device_response: Callable[[], bool]) -> bool:
        dialog = BitBox02PairingDialog(code, device_response)
        dialog.show()
        # render the window since the next operation is blocking
        while True:
            QCoreApplication.processEvents()
            if dialog.painted:
                break
            time.sleep(0.1)
        if not device_response():
            return False
        dialog.enable_buttons()
        dialog.exec_()
        return dialog.result() == QDialog.Accepted

    def attestation_check(self, result: bool) -> None:
        if not result:
            QMessageBox.warning(
                None,
                "BitBox02 attestation check",
                "BitBox02 attestation check failed. Your BitBox02 might not be genuine. Please contact support@shiftcrypto.ch if the problem persists.",
            )

class HWIQt(QMainWindow):
    def __init__(self, passphrase='', chain=Chain.MAIN):
        super(HWIQt, self).__init__()
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)
        self.setWindowTitle('HWI Qt')

        self.devices = []
        self.client = None
        self.device_info = {}
        self.passphrase = passphrase
        self.chain = chain
        self.current_dialog = None
        self.getkeypool_opts = {
            'start': 0,
            'end': 1000,
            'account': 0,
            'internal': False,
            'keypool': True,
            'addrtype': AddressType.SH_WIT,
            'path': None,
            'account_used': True
        }

        self.ui.enumerate_refresh_button.clicked.connect(self.refresh_clicked)
        self.ui.setpass_button.clicked.connect(self.show_setpassphrasedialog)
        self.ui.sendpin_button.clicked.connect(lambda: self.show_sendpindialog(prompt_pin=True))
        self.ui.getxpub_button.clicked.connect(self.show_getxpubdialog)
        self.ui.signtx_button.clicked.connect(self.show_signpsbtdialog)
        self.ui.signmsg_button.clicked.connect(self.show_signmessagedialog)
        self.ui.display_addr_button.clicked.connect(self.show_displayaddressdialog)
        self.ui.getkeypool_opts_button.clicked.connect(self.show_getkeypooloptionsdialog)
        self.ui.toggle_passphrase_button.clicked.connect(self.toggle_passphrase)

        self.ui.enumerate_combobox.currentIndexChanged.connect(self.get_client_and_device_info)

    def clear_info(self):
        self.ui.getxpub_button.setEnabled(False)
        self.ui.signtx_button.setEnabled(False)
        self.ui.signmsg_button.setEnabled(False)
        self.ui.display_addr_button.setEnabled(False)
        self.ui.getkeypool_opts_button.setEnabled(False)
        self.ui.toggle_passphrase_button.setEnabled(False)
        self.ui.keypool_textedit.clear()
        self.ui.desc_textedit.clear()

    @Slot()
    def refresh_clicked(self):
        if self.client:
            self.client.close()
            self.client = None

        self.devices = commands.enumerate(self.passphrase)
        self.ui.enumerate_combobox.currentIndexChanged.disconnect()
        self.ui.enumerate_combobox.clear()
        self.ui.enumerate_combobox.addItem('')
        for dev in self.devices:
            fingerprint = 'none'
            if 'fingerprint' in dev:
                fingerprint = dev['fingerprint']
            dev_str = '{} fingerprint:{} path:{}'.format(dev['model'], fingerprint, dev['path'])
            self.ui.enumerate_combobox.addItem(dev_str)
        self.ui.enumerate_combobox.currentIndexChanged.connect(self.get_client_and_device_info)
        self.clear_info()

    @Slot()
    def show_setpassphrasedialog(self):
        self.current_dialog = SetPassphraseDialog()
        self.current_dialog.accepted.connect(self.setpassphrasedialog_accepted)
        self.current_dialog.exec_()

    @Slot()
    def setpassphrasedialog_accepted(self):
        self.passphrase = self.current_dialog.ui.passphrase_lineedit.text()
        self.current_dialog = None

    @Slot()
    def get_client_and_device_info(self, index):
        self.ui.sendpin_button.setEnabled(False)
        if index == 0:
            self.clear_info()
            return

        self.ui.getxpub_button.setEnabled(True)
        self.ui.signtx_button.setEnabled(True)
        self.ui.display_addr_button.setEnabled(True)
        self.ui.getkeypool_opts_button.setEnabled(True)

        # Get the client
        self.device_info = self.devices[index - 1]
        self.client = commands.get_client(self.device_info['model'], self.device_info['path'], self.passphrase)
        self.client.chain = self.chain

        if self.device_info['type'] == 'bitbox02':
            self.client.set_noise_config(BitBox02NoiseConfig())

        self.ui.setpass_button.setEnabled(self.device_info['type'] != 'bitbox02')
        self.ui.signmsg_button.setEnabled(self.device_info['type'] != 'bitbox02')
        self.ui.toggle_passphrase_button.setEnabled(self.device_info['type'] in ('trezor', 'keepkey', 'bitbox02', ))

        self.get_device_info()

    def get_device_info(self):
        # Enable the sendpin button if it's a trezor and it needs it
        if self.device_info['needs_pin_sent']:
            self.ui.sendpin_button.setEnabled(True)
            self.clear_info()
            return
        else:
            self.ui.sendpin_button.setEnabled(False)

        # If it isn't initialized, show an error but don't do anything
        if 'code' in self.device_info and self.device_info['code'] == DEVICE_NOT_INITIALIZED:
            self.clear_info()
            QMessageBox.information(None, "Not initialized yet", 'Device is not initalized yet')
            return

        # do getkeypool and getdescriptors
        keypool = do_command(commands.getkeypool, self.client,
                             None if self.getkeypool_opts['account_used'] else self.getkeypool_opts['path'],
                             self.getkeypool_opts['start'],
                             self.getkeypool_opts['end'],
                             self.getkeypool_opts['internal'],
                             self.getkeypool_opts['keypool'],
                             self.getkeypool_opts['account'],
                             self.getkeypool_opts['addrtype'])
        descriptors = do_command(commands.getdescriptors, self.client, self.getkeypool_opts['account'])

        self.ui.keypool_textedit.setPlainText(json.dumps(keypool, indent=2))
        self.ui.desc_textedit.setPlainText(json.dumps(descriptors, indent=2))

    @Slot()
    def show_sendpindialog(self, prompt_pin=True):
        self.current_dialog = SendPinDialog(self.client, prompt_pin)
        self.current_dialog.pin_sent_success.connect(self.sendpindialog_accepted)
        self.current_dialog.exec_()

    @Slot()
    def sendpindialog_accepted(self):
        self.current_dialog = None

        curr_index = self.ui.enumerate_combobox.currentIndex()
        self.refresh_clicked()
        self.ui.enumerate_combobox.setCurrentIndex(curr_index)

    @Slot()
    def show_getxpubdialog(self):
        self.current_dialog = GetXpubDialog(self.client)
        self.current_dialog.exec_()

    @Slot()
    def show_signpsbtdialog(self):
        self.current_dialog = SignPSBTDialog(self.client)
        self.current_dialog.exec_()

    @Slot()
    def show_signmessagedialog(self):
        self.current_dialog = SignMessageDialog(self.client)
        self.current_dialog.exec_()

    @Slot()
    def show_displayaddressdialog(self):
        self.current_dialog = DisplayAddressDialog(self.client)
        self.current_dialog.exec_()

    @Slot()
    def show_getkeypooloptionsdialog(self):
        self.current_dialog = GetKeypoolOptionsDialog(self.getkeypool_opts)
        self.current_dialog.accepted.connect(self.getkeypooloptionsdialog_accepted)
        self.current_dialog.exec_()

    @Slot()
    def getkeypooloptionsdialog_accepted(self):
        self.getkeypool_opts['start'] = self.current_dialog.ui.start_spinbox.value()
        self.getkeypool_opts['end'] = self.current_dialog.ui.end_spinbox.value()
        self.getkeypool_opts['internal'] = self.current_dialog.ui.internal_checkbox.isChecked()
        self.getkeypool_opts['keypool'] = self.current_dialog.ui.keypool_checkbox.isChecked()
        self.getkeypool_opts['addrtype'] = AddressType.LEGACY
        if self.current_dialog.ui.sh_wpkh_radio.isChecked():
            self.getkeypool_opts['addrtype'] = AddressType.SH_WIT
        if self.current_dialog.ui.wpkh_radio.isChecked():
            self.getkeypool_opts['addrtype'] = AddressType.WIT
        if self.current_dialog.ui.pkh_radio.isChecked():
            self.getkeypool_opts['addrtype'] = AddressType.LEGACY
        if self.current_dialog.ui.account_radio.isChecked():
            self.getkeypool_opts['account'] = self.current_dialog.ui.account_spinbox.value()
            self.getkeypool_opts['account_used'] = True
        else:
            self.getkeypool_opts['path'] = self.current_dialog.ui.path_lineedit.text()
            self.getkeypool_opts['account_used'] = False
        self.current_dialog = None
        self.get_device_info()

    @Slot()
    def toggle_passphrase(self):
        do_command(commands.toggle_passphrase, self.client)
        if self.device_info['model'] == "keepkey":
            self.show_sendpindialog(prompt_pin=False)

def process_gui_commands(cli_args):
    parser = HWIArgumentParser(description='Hardware Wallet Interface Qt, version {}.\nInteractively access and send commands to a hardware wallet device with a GUI. Responses are in JSON format.'.format(__version__))
    parser.add_argument('--password', '-p', help='Device password if it has one (e.g. DigitalBitbox)', default='')
    parser.add_argument('--chain', help='Select chain to work with', type=Chain.argparse, choices=list(Chain), default=Chain.MAIN)
    parser.add_argument('--debug', help='Print debug statements', action='store_true')
    parser.add_argument('--version', action='version', version='%(prog)s {}'.format(__version__))

    # Parse arguments again for anything entered over stdin
    args = parser.parse_args(cli_args)

    result = {}

    # Setup debug logging
    logging.basicConfig(level=logging.DEBUG if args.debug else logging.WARNING)

    # Qt setup
    app = QApplication()

    window = HWIQt(args.password, args.chain)

    window.refresh_clicked()

    window.show()
    ret = app.exec_()
    result = {'success': ret == 0}

    return result

def main():
    result = process_gui_commands(sys.argv[1:])
    print(json.dumps(result))
